import dsgp4
import numpy as np
import random
import torch
import unittest

error_string="Error: deep space propagation not supported (yet). The provided satellite has \
an orbital period above 225 minutes. If you want to let us know you need it or you want to \
contribute to implement it, open a PR or raise an issue at: https://github.com/esa/dSGP4."

class UtilTestCase(unittest.TestCase):
    def test_mldsgp4_single_tles(self):
        lines=file.splitlines()
        #I randomly select 50 indexes out of 500 satellites
        indexes=list(range(1,len(lines),3))
        tles=[]
        for i in indexes:
            data=[]
            data.append(lines[i])
            data.append(lines[i+1])
            data.append(lines[i+2])
            tles.append(dsgp4.tle.TLE(data))

        #let's load the ML-dSGP4 model without input/output correction
        #this should correspond to just SGP4
        ml_dsgp4=dsgp4.mldsgp4(hidden_size=35,
                               input_correction=0.,
                               output_correction=0.,
                               normalization_R=6958.137, 
                               normalization_V=7.947155867983262)
        #random times:
        tsinces=torch.rand(len(tles),)*10000
        for tle in tles:
            #we first propagate with the ML-dSGP4:
            try:
                #we use torch.no_grad() to avoid keeping track of the derivatives:
                with torch.no_grad():    
                    states_mldsgp4_out=ml_dsgp4(tle,tsinces)
                #let's unnormalize the output:
                states_mldsgp4_out[:,:3]*=ml_dsgp4.normalization_R
                states_mldsgp4_out[:,3:]*=ml_dsgp4.normalization_V
            except Exception as e:
                self.assertTrue((str(e).split()==error_string.split()))
            #now with the SGP4:
            try:
                states_dsgp4_out=dsgp4.propagate(tle, tsinces, initialized=False)
            except Exception as e:
                self.assertTrue((str(e).split()==error_string.split()))
            #testing the results:
            self.assertTrue(np.allclose(states_mldsgp4_out.detach().numpy().reshape(-1,2,3),states_dsgp4_out.detach().numpy(),atol=1e-13))

    def test_mldsgp4_batch_tles(self):
        lines=file.splitlines()
        #I randomly select 50 indexes out of 500 satellites
        indexes=list(range(1,len(lines),3))
        tles=[]
        for i in indexes:
            data=[]
            data.append(lines[i])
            data.append(lines[i+1])
            data.append(lines[i+2])
            tles.append(dsgp4.tle.TLE(data))
        tles=tles[:20]
        #let's load the ML-dSGP4 model without input/output correction
        #this should correspond to just SGP4
        ml_dsgp4=dsgp4.mldsgp4(hidden_size=35,
                               input_correction=0.,
                               output_correction=0.,
                               normalization_R=6958.137, 
                               normalization_V=7.947155867983262)
        #we now create a batch of TLE to later propagate it
        tles_=[]
        for tle in tles:
            tles_+=[tle]*100
        tsinces = torch.cat([torch.linspace(0,24*60,100)]*len(tles))

        #we first propagate with the ML-dSGP4 and we use torch.no_grad() to avoid keeping track of the derivatives:
        with torch.no_grad():    
            states_mldsgp4_out=ml_dsgp4(tles_,tsinces)
        #let's unnormalize the output:
        states_mldsgp4_out[:,:3]*=ml_dsgp4.normalization_R
        states_mldsgp4_out[:,3:]*=ml_dsgp4.normalization_V
        states_dsgp4_out=dsgp4.propagate_batch(tles_, tsinces, initialized=False)
        #testing the results:
        self.assertTrue(np.allclose(states_mldsgp4_out.detach().numpy().reshape(-1,2,3),states_dsgp4_out.detach().numpy(),atol=1e-12))


file="""
0 COSMOS 2251 DEB
1 34613U 93036WJ  22068.72863866  .00001082  00000-0  53831-3 0  9996
2 34613  73.9681 223.5130 0129564 354.9799  16.8811 14.13981619669720
0 COSMOS 2251 DEB
1 34614U 93036WK  22068.63183764  .00004007  00000-0  24848-2 0  9994
2 34614  73.8895 226.7217 0273044 126.6005   3.7416 13.88618176655351
0 COSMOS 2251 DEB
1 34615U 93036WL  22069.18146051  .00000335  00000-0  33390-3 0  9997
2 34615  74.0062 331.1428 0409982 159.0926 327.4818 13.49468644642707
0 COSMOS 2251 DEB
1 34616U 93036WM  22068.66873903  .00079618  00000-0  50636-2 0  9994
2 34616  74.0152 264.4423 0009130 189.7533 287.2375 15.08363272693295
0 COSMOS 2251 DEB
1 34619U 93036WQ  22068.93716744  .00000494  00000-0  10379-3 0  9993
2 34619  74.0476 306.5313 0101351  34.0169 326.7438 14.59259854692280
0 COSMOS 2251 DEB
1 34620U 93036WR  22068.72189948  .00003328  00000-0  14138-2 0  9997
2 34620  74.0686 203.0250 0123920 221.3081 137.8643 14.21617239671578
0 COSMOS 2251 DEB
1 34622U 93036WT  22068.83487469  .00000686  00000-0  34723-3 0  9997
2 34622  74.0424 234.4368 0119529 285.2354 141.9703 14.14417339669908
0 COSMOS 2251 DEB
1 34626U 93036WX  22069.12334337  .00001420  00000-0  33666-3 0  9991
2 34626  74.0573  68.4773 0038853 158.9112 271.0263 14.54489517690888
0 COSMOS 2251 DEB
1 34629U 93036XA  22068.48260093  .00014277  00000-0  11303-2 0  9996
2 34629  73.7784 227.2057 0054198 261.9499  97.5534 14.99449156702738
0 COSMOS 2251 DEB
1 34634U 93036XF  22068.86553663  .00000502  00000-0  13689-3 0  9998
2 34634  74.0521 100.2999 0051675 197.9671 217.4335 14.49249934688427
0 COSMOS 2251 DEB
1 34635U 93036XG  22068.70033788  .00001820  00000-0  48750-3 0  9994
2 34635  74.0537 153.8664 0046265 215.2520 267.6548 14.48319810532212
0 COSMOS 2251 DEB
1 34637U 93036XJ  22068.72107742  .00001738  00000-0  47830-3 0  9996
2 34637  74.0547 166.9932 0066667 226.8243 251.2590 14.46396447685850
0 COSMOS 2251 DEB
1 34638U 93036XK  22068.91656580  .00003294  00000-0  73037-3 0  9994
2 34638  74.0435 106.0047 0008556 157.1245 261.9065 14.57500207687535
0 COSMOS 2251 DEB
1 34639U 93036XL  22068.68400142  .00004423  00000-0  10283-2 0  9998
2 34639  74.0334 147.5146 0043510 201.4877 286.4544 14.54631374685390
0 COSMOS 2251 DEB
1 34641U 93036XN  22068.91549830  .00001103  00000-0  36055-3 0  9993
2 34641  73.8436 149.8823 0020840 334.5142 146.7654 14.39691670684166
0 COSMOS 2251 DEB
1 34642U 93036XP  22069.01466616  .00000998  00000-0  66107-3 0  9998
2 34642  74.0075 181.2175 0227764 302.2655 174.4725 13.90985389661918
0 IRIDIUM 33 DEB
1 34643U 97051JU  22068.68681555  .00006654  00000-0  14774-2 0  9990
2 34643  86.3796 251.6356 0035333 160.5044 199.7525 14.56546284683433
0 IRIDIUM 33 DEB
1 34648U 97051JZ  22069.10136961  .00002815  00000-0  18997-2 0  9995
2 34648  86.4076 101.5871 0230215 199.1784 160.0616 13.88148366658369
0 IRIDIUM 33 DEB
1 34651U 97051KC  22068.20655725  .00001931  00000-0  10436-2 0  9991
2 34651  86.2956 352.1986 0188805 205.0191 154.1750 14.03506581664900
0 IRIDIUM 33 DEB
1 34652U 97051KD  22068.71059411  .00001745  00000-0  46623-3 0  9999
2 34652  86.3381 222.4216 0036042 230.3439 140.9446 14.47457461686011
0 NAVSTAR 63 (USA 203)
1 34661U 09014A   22068.49766742  .00000025  00000-0  00000-0 0  9994
2 34661  55.1247 273.3242 0121622  55.4323 306.7902  2.00559615 94985
0 SL-12 DEB
1 34667U 91025L   22069.09518701  .00000568  00000-0  86705-2 0  9990
2 34667  65.8889 280.5501 5458247  92.4487 357.8727  4.35288851205462
0 COSMOS 2251 DEB
1 34671U 93036XR  22068.82793444  .00001830  00000-0  53720-3 0  9993
2 34671  74.0351 206.7640 0021166 263.1629 222.0589 14.44403303683442
0 COSMOS 2251 DEB
1 34672U 93036XS  22068.41156390  .00000300  00000-0  11768-3 0  9999
2 34672  74.0352 303.6683 0025348 351.0783 182.6152 14.33502764454679
0 COSMOS 2251 DEB
1 34674U 93036XU  22068.76822552  .00003298  00000-0  58970-3 0  9994
2 34674  73.9845 332.0633 0111666 257.4278 101.4382 14.63317013690540
0 COSMOS 2251 DEB
1 34675U 93036XV  22068.85898012  .00003276  00000-0  76511-3 0  9996
2 34675  74.0348 137.3326 0032168 196.8713 163.1383 14.54798835686024
0 COSMOS 2251 DEB
1 34678U 93036XY  22068.85041662  .00000025  00000-0  19743-4 0  9994
2 34678  74.0395 342.6753 0025400   2.9075  58.8849 14.29731025677424
0 COSMOS 2251 DEB
1 34679U 93036XZ  22068.83834898  .00005961  00000-0  93844-3 0  9991
2 34679  74.0089 277.5975 0042065  22.4234 337.8764 14.72112204693342
0 COSMOS 2251 DEB
1 34680U 93036YA  22068.46794456  .00001710  00000-0  49899-3 0  9996
2 34680  74.0494 193.3378 0011924 263.2830 168.3865 14.44826463683945
0 COSMOS 2251 DEB
1 34681U 93036YB  22068.71717113  .00002091  00000-0  46957-3 0  9995
2 34681  74.0746  54.7516 0040284 121.5808  51.1250 14.56677786691279
0 COSMOS 2251 DEB
1 34682U 93036YC  22068.91068493  .00001075  00000-0  30263-3 0  9992
2 34682  74.0441 149.1970 0024657 260.9363 222.5317 14.46901496687499
0 COSMOS 2251 DEB
1 34683U 93036YD  22068.72708171  .00001739  00000-0  48845-3 0  9998
2 34683  74.0362 167.4910 0028868 257.3547 222.0571 14.46457811683851
0 COSMOS 2251 DEB
1 34685U 93036YF  22069.15636324  .00001324  00000-0  40077-3 0  9997
2 34685  74.0353 204.2239 0019721 280.8056 253.8297 14.43205135686892
0 COSMOS 2251 DEB
1 34687U 93036YH  22069.09795462  .00009616  00000-0  67555-3 0  9993
2 34687  73.9226 211.9798 0083273  41.8298  74.6081 15.02397782707016
0 COSMOS 2251 DEB
1 34689U 93036YK  22068.39672330  .00002967  00000-0  20466-2 0  9996
2 34689  74.1011  41.0080 0284977  81.2570  69.9302 13.81824768652313
0 IRIDIUM 33 DEB
1 34690U 97051KM  22068.83364907  .00001478  00000-0  43019-3 0  9992
2 34690  86.3819 266.9213 0008297 294.2361 126.7797 14.43629843682055
0 IRIDIUM 33 DEB
1 34693U 97051KQ  22068.90768167  .00003071  00000-0  68981-3 0  9995
2 34693  86.3422 208.0805 0027161 123.3364   0.2315 14.56008043686657
0 IRIDIUM 33 DEB
1 34696U 97051KT  22069.01427476  .00002379  00000-0  10771-2 0  9999
2 34696  86.2042 272.0186 0139501 313.0537 200.2712 14.16615764670478
0 IRIDIUM 33 DEB
1 34702U 97051KZ  22068.50074806  .00003193  00000-0  62095-3 0  9997
2 34702  86.3131 172.3314 0038928 359.9862   0.1341 14.62357707691099
0 IRIDIUM 33 DEB
1 34705U 97051LC  22068.85690611  .00013199  00000-0  16888-2 0  9991
2 34705  86.4603 285.1272 0073387   0.1236 359.9989 14.79345575689044
0 IRIDIUM 33 DEB
1 34706U 97051LD  22068.93861435  .00037695  00000-0  23337-2 0  9999
2 34706  86.3021 126.0688 0018169  51.3179  68.4484 15.09494923698029
0 EUTE 10A (EUTE W2A)
1 34710U 09016A   22068.90577198  .00000065  00000-0  00000-0 0  9996
2 34710   0.0616   1.6065 0005540 350.7126 151.2827  1.00272168 25613
0 BREEZE-M R/B
1 34711U 09016B   22068.65983842 -.00000007  00000-0  00000-0 0  9993
2 34711  20.5652 253.1520 5936734 183.9473 167.5621  2.20367276112140
0 BREEZE-M DEB (TANK)
1 34712U 09016C   22068.68120304  .00003542  00000-0  51437-3 0  9992
2 34712  46.0467 243.9004 5042943 233.3680  69.0197  5.55861109255565
0 WGS F2 (USA 204)
1 34713U 09017A   22069.13046942  .00000065  00000-0  00000+0 0  9993
2 34713   0.0173 160.5373 0001922  79.0554  32.6592  1.00270811 47299
0 ATLAS 5 CENTAUR R/B
1 34714U 09017B   22066.24526347  .00001011  00000-0  35466-2 0  9994
2 34714  16.8402 252.9671 8246482 246.5776  13.9459  1.16266407 54288
0 COSMOS 2251 DEB
1 34721U 93036YP  22069.05735669  .00000876  00000-0  59488-3 0  9996
2 34721  73.9246 170.2277 0238990  22.0153 151.6200 13.88659436660347
0 COSMOS 2251 DEB
1 34722U 93036YQ  22068.80660747  .00000515  00000-0  28055-3 0  9999
2 34722  73.8871 252.3946 0178278  86.0938 287.8838 14.06747396668946
0 COSMOS 2251 DEB
1 34723U 93036YR  22067.64909525  .12226285  19405-5  60915-2 0  9992
2 34723  73.9258 293.0276 0008565 260.5674  99.4607 16.20633215700335
0 COSMOS 2251 DEB
1 34724U 93036YS  22068.95171465  .00009396  00000-0  44780-2 0  9999
2 34724  74.0463 170.8083 0235314 247.2857 110.3254 14.05916928659456
0 COSMOS 2251 DEB
1 34725U 93036YT  22068.47891361  .00017653  00000-0  28047-2 0  9994
2 34725  74.2020 197.1756 0038030 342.3895  17.5956 14.71586483686697
0 COSMOS 2251 DEB
1 34726U 93036YU  22068.76244980  .00000861  00000-0  22039-3 0  9995
2 34726  74.0287  66.8109 0051070 200.3154 159.5977 14.51169034685242
0 COSMOS 2251 DEB
1 34728U 93036YW  22068.88428617  .00001155  00000-0  26121-3 0  9994
2 34728  74.0252   5.3476 0054228 149.2994 211.1362 14.56491785687197
0 COSMOS 2251 DEB
1 34729U 93036YX  22069.15419014  .00001673  00000-0  42481-3 0  9999
2 34729  74.0282 112.0535 0030001 218.5248 208.0466 14.51309990684208
0 COSMOS 2251 DEB
1 34732U 93036ZA  22068.45347061  .00000237  00000-0  74752-4 0  9999
2 34732  74.0447 136.1794 0042142 270.2175 101.0719 14.45188364682789
0 COSMOS 2251 DEB
1 34735U 93036ZD  22068.29627898  .00009433  00000-0  12823-2 0  9999
2 34735  74.0294 282.7872 0047738 336.6650  23.2357 14.78105639692603
0 COSMOS 2251 DEB
1 34736U 93036ZE  22068.79062323  .00003254  00000-0  86845-3 0  9992
2 34736  74.0360 182.9041 0017208 226.1241 251.9473 14.48715830682587
0 COSMOS 2251 DEB
1 34737U 93036ZF  22068.83700856  .00009138  00000-0  16175-2 0  9993
2 34737  74.0362  85.5052 0022213   0.9773  53.0536 14.67285749686026
0 COSMOS 2251 DEB
1 34738U 93036ZG  22068.77818821  .00000488  00000-0  15521-3 0  9993
2 34738  74.0189 180.3641 0050867 297.8068 189.0747 14.41834665680300
0 COSMOS 2251 DEB
1 34739U 93036ZH  22068.34657073  .00010594  00000-0  13328-2 0  9998
2 34739  74.0264 182.5803 0055026 262.4966 159.5514 14.81060712696338
0 COSMOS 2251 DEB
1 34741U 93036ZK  22069.14364918  .00046838  00000-0  20549-2 0  9996
2 34741  73.9931 346.1735 0064179  66.4127 294.3792 15.19854917702538
0 COSMOS 2251 DEB
1 34744U 93036ZN  22068.82408244  .00005846  00000-0  89039-3 0  9998
2 34744  74.0081 265.8069 0036861  14.7410 357.0262 14.73695161692141
0 COSMOS 2251 DEB
1 34745U 93036ZP  22069.16749933  .00044130  00000-0  26656-2 0  9994
2 34745  74.0130  89.1565 0012838 237.2251 122.7705 15.10411539698892
0 COSMOS 2251 DEB
1 34746U 93036ZQ  22068.87299352  .00007025  00000-0  12229-2 0  9990
2 34746  74.0281 359.6835 0019520  79.2341 336.2123 14.68125772691425
0 COSMOS 2251 DEB
1 34747U 93036ZR  22068.83221079  .00015800  00000-0  13067-2 0  9999
2 34747  73.9864 355.9292 0044633 144.7464 215.6688 14.98129691701788
0 COSMOS 2251 DEB
1 34750U 93036ZU  22068.69062684  .00001609  00000-0  47293-3 0  9993
2 34750  74.1154 252.4121 0053097 217.5174 268.5978 14.43757078683076
0 COSMOS 2251 DEB
1 34751U 93036ZV  22068.85636538  .00005175  00000-0  13522-2 0  9996
2 34751  74.0067 215.2526 0035843 235.1604 248.2185 14.49267877683702
0 COSMOS 2251 DEB
1 34752U 93036ZW  22069.04619078  .00001738  00000-0  27856-3 0  9996
2 34752  74.0239 192.7702 0089161 313.3233  46.0528 14.69940591695204
0 COSMOS 2251 DEB
1 34754U 93036ZY  22068.90887756  .00000301  00000-0  10871-3 0  9996
2 34754  74.0301 241.9607 0030919 333.0041 149.7006 14.37578019683745
0 COSMOS 2251 DEB
1 34755U 93036ZZ  22068.96073263  .00000153  00000-0  16973-3 0  9997
2 34755  73.9295 329.8585 0420789 244.4607 124.2380 13.46911975640060
0 COSMOS 2251 DEB
1 34756U 93036AAA 22068.69017199  .00000626  00000-0  54737-3 0  9999
2 34756  74.0048 191.9284 0361339 354.7345   5.0006 13.61679417646855
0 COSMOS 2251 DEB
1 34758U 93036AAC 22068.80215110  .00001416  00000-0  12909-2 0  9999
2 34758  74.1098 341.4888 0398159  61.1150 302.9298 13.53423274642934
0 COSMOS 2251 DEB
1 34759U 93036AAD 22068.85701524  .00015822  00000-0  22165-2 0  9995
2 34759  74.0166   4.0855 0010827   5.3999 354.7291 14.77482314693559
0 COSMOS 2251 DEB
1 34760U 93036AAE 22069.19752550  .00000866  00000-0  43707-3 0  9998
2 34760  74.0644 223.2292 0102496 251.3660 284.0339 14.15081754672644
0 COSMOS 2251 DEB
1 34761U 93036AAF 22068.52064119  .00000971  00000-0  37594-3 0  9996
2 34761  74.0333 358.0392 0041935   5.3956 354.7637 14.30896942680097
0 COSMOS 2251 DEB
1 34763U 93036AAH 22068.78065034  .00008832  00000-0  18512-2 0  9995
2 34763  74.1730 313.6296 0064750 119.4211 308.0251 14.58493953683870
0 IRIDIUM 33 DEB
1 34764U 97051LH  22068.57543325  .00054250  00000-0  39325-2 0  9998
2 34764  86.3538 178.2573 0005954 260.8955  99.1605 15.03689251694515
0 IRIDIUM 33 DEB
1 34765U 97051LJ  22068.78051463  .00000657  00000-0  22352-3 0  9994
2 34765  86.3366 258.8957 0018995  46.0804 323.8526 14.35014857678780
0 IRIDIUM 33 DEB
1 34773U 97051LS  22068.71126542  .00004841  00000-0  10962-2 0  9999
2 34773  86.4072 264.8974 0017105  72.0967 288.2104 14.55898486688104
0 IRIDIUM 33 DEB
1 34774U 97051LT  22069.05934824  .00004012  00000-0  79091-3 0  9997
2 34774  86.4186 259.0488 0021525 326.6878 158.1491 14.62151183690726
0 IRIDIUM 33 DEB
1 34775U 97051LU  22068.81064666  .00002627  00000-0  74056-3 0  9995
2 34775  86.3753 269.4146 0026287 281.8282  89.7599 14.45272552682299
0 COMPASS G2
1 34779U 09018A   22068.02701842 -.00000151  00000-0  00000-0 0  9998
2 34779   8.2706  50.6554 0319604 249.7419 106.8765  0.94516854 47397
0 CZ-3C R/B
1 34780U 09018B   22069.16415195  .00005617  00000-0  24655-2 0  9994
2 34780  20.7006 321.7320 7131875  39.9300 354.9073  2.45578427112046
0 COSMOS 2251 DEB
1 34784U 93036AAM 22068.36638343  .00000887  00000-0  22329-3 0  9995
2 34784  74.0090  52.3058 0060054 197.0015 162.9136 14.51648502685406
0 COSMOS 2251 DEB
1 34787U 93036AAQ 22069.09891571  .00006428  00000-0  10002-2 0  9994
2 34787  74.0161 309.5714 0013326  73.4927 286.7708 14.73138766692625
0 COSMOS 2251 DEB
1 34788U 93036AAR 22068.94122777  .00003389  00000-0  29271-3 0  9995
2 34788  73.9903 280.8094 0134553  31.4587  33.2058 14.90884651707015
0 COSMOS 2251 DEB
1 34789U 93036AAS 22068.79813514  .00004225  00000-0  14561-2 0  9990
2 34789  74.2048 192.1639 0103795  18.1297 106.5324 14.33192302674468
0 COSMOS 2251 DEB
1 34790U 93036AAT 22068.77443678  .00000752  00000-0  35584-3 0  9990
2 34790  74.0573 182.7815 0109231 218.6289 269.6003 14.18323488670567
0 COSMOS 2251 DEB
1 34791U 93036AAU 22068.74233478  .00008295  00000-0  85591-3 0  9996
2 34791  74.0186  83.2071 0046207 213.6720 146.1524 14.89529078701119
0 COSMOS 2251 DEB
1 34792U 93036AAV 22068.52062462  .00001556  00000-0  48324-3 0  9992
2 34792  74.0244 222.0367 0016818 284.6326 138.2886 14.41827284677664
0 COSMOS 2251 DEB
1 34793U 93036AAW 22068.80776917  .00000342  00000-0  12642-3 0  9997
2 34793  74.0020 260.9066 0053385  18.5431 353.3504 14.35371548675985
0 COSMOS 2251 DEB
1 34796U 93036AAZ 22068.93184426  .00000768  00000-0  23839-3 0  9994
2 34796  74.0333 248.2516 0116421 327.7024 155.9459 14.39028736682118
0 COSMOS 2251 DEB
1 34797U 93036ABA 22068.80523530  .00001512  00000-0  62383-3 0  9999
2 34797  73.9824  95.8766 0092337 161.8617 198.5862 14.25350974677447
0 RISAT 2
1 34807U 09019A   22069.20015293  .00039560  00000-0  69476-3 0  9990
2 34807  41.2102 146.3902 0010166  13.1639  10.3869 15.49708764724344
0 BLOCK DM-SL R/B
1 34811U 09020B   22068.85335389 -.00000141  00000-0  00000-0 0  9997
2 34811   2.0707  53.4388 4737112 332.3433 217.9075  1.81487904 85390
"""