import dsgp4
import numpy as np
import random
import sgp4
import sgp4.io
import sgp4.earth_gravity
import torch
import unittest

error_string="Error: deep space propagation not supported (yet). The provided satellite has \
an orbital period above 225 minutes. If you want to let us know you need it or you want to \
contribute to implement it, open a PR or raise an issue at: https://github.com/esa/dSGP4."

class UtilTestCase(unittest.TestCase):
    def test_sgp4(self):
        lines=file.splitlines()
        #I randomly select 50 indexes out of 500 satellites
        indexes=random.sample(list(range(1,len(lines),3)), 50)
        tles=[]
        for i in indexes:
            data=[]
            data.append(lines[i])
            data.append(lines[i+1])
            data.append(lines[i+2])
            tles.append(dsgp4.tle.TLE(data))
        for tle_sat in tles:
            satrec=sgp4.io.twoline2rv(tle_sat.line1,tle_sat.line2, whichconst=sgp4.earth_gravity.wgs84)
            sgp4.propagation.sgp4init(whichconst=sgp4.earth_gravity.wgs84,
                                opsmode='i',
                                satn=tles[0].satellite_catalog_number,
                                epoch=(satrec.jdsatepoch+satrec.jdsatepochF)-2433281.5,
                                xbstar=satrec.bstar,
                                xndot=satrec.ndot,
                                xnddot=satrec.nddot,
                                xecco=satrec.ecco,
                                xargpo=satrec.argpo,
                                xinclo=satrec.inclo,
                                xmo=satrec.mo,
                                xno_kozai=satrec.no_kozai,
                                xnodeo=satrec.nodeo,
                                satrec=satrec)
            try:
                dsgp4.initialize_tle(tle_sat, gravity_constant_name="wgs-84");
                for _ in range(50):
                    tsince=random.uniform(-100.,100.)
                    satrec_state=sgp4.propagation.sgp4(satrec, tsince)
                    tle_sat_state=dsgp4.sgp4(tle_sat, tsince*torch.ones(1,1))
                    self.assertAlmostEqual(satrec_state[0][0], float(tle_sat_state[0][0]))
                    self.assertAlmostEqual(satrec_state[0][1], float(tle_sat_state[0][1]))
                    self.assertAlmostEqual(satrec_state[0][2], float(tle_sat_state[0][2]))
                    self.assertAlmostEqual(satrec_state[1][0], float(tle_sat_state[1][0]))
                    self.assertAlmostEqual(satrec_state[1][1], float(tle_sat_state[1][1]))
                    self.assertAlmostEqual(satrec_state[1][2], float(tle_sat_state[1][2]))
            except Exception as e:
                self.assertTrue((str(e).split()==error_string.split()))

    def test_batch_propagation(self):
        lines=file.splitlines()
        #I randomly select 50 indexes out of 500 satellites
        indexes=random.sample(list(range(1,len(lines),3)), 50)
        tles=[]
        for i in indexes:
            data=[]
            data.append(lines[i])
            data.append(lines[i+1])
            data.append(lines[i+2])
            tles.append(dsgp4.tle.TLE(data))

        #I filter out deep space and error cases:
        tles_filtered=[]
        for tle_satellite in tles:
            try:
                dsgp4.initialize_tle(tle_satellite, gravity_constant_name="wgs-84");
                if tle_satellite._error==0:
                    tles_filtered.append(tle_satellite)
            except Exception as e:
                self.assertTrue((str(e).split()==error_string.split()))

        for tle in tles_filtered:
            satrec=sgp4.io.twoline2rv(tle.line1,tle.line2, whichconst=sgp4.earth_gravity.wgs84)
            tsinces=torch.rand(100,)*10
            satrec_states=np.stack([np.array(sgp4.propagation.sgp4(satrec, t)) for t in tsinces.detach().numpy()])
            self.assertTrue(np.allclose(dsgp4.sgp4(tle,tsinces).detach().numpy(),satrec_states))
        
    def test_spacetrack(self):
        #here we compare the output with the output of the spacetrack API (attached at the end of this file), for 78 different TLEs, propagated at 5 different times each
        #for more details on the official SGP4 implementation, see: https://www.space-track.org/documentation#/sgp4
        lines=out_file.splitlines()#see end of this file
        outs=[]
        start=19#initial rows are infos about the propagation
        for i in np.arange(0,len(lines)-20,10): 
            outs.append(lines[start+i:start+i+5])
        #we gather the results:
        space_track_results=[]
        for d in outs:
            matrix = np.array([[float(val) for val in row.split()] for row in d])
            space_track_results.append(matrix)
        #we now collect the matrix of: N_tles x N_tsinces x N_columns
        #where the columns are: tsince, x, y, z, xdot, ydot, zdot (in km and km/s)
        space_track_results=np.stack(space_track_results)
        #we now propagate with dSGP4 the same TLEs, at the same times:
        #I read the TLEs
        lines=file.splitlines()[:int(78*3)+1]#I only load the first 78 TLEs
        print(len(lines))
        indexes=list(range(1,len(lines),3))
        #I now construct the TLE objects and initialize the dSGP4 propagator
        tles=[]
        for i in indexes:
            data=[]
            data.append(lines[i])
            data.append(lines[i+1])
            data.append(lines[i+2])
            tles.append(dsgp4.tle.TLE(data))
            dsgp4.initialize_tle(tles[-1], gravity_constant_name="wgs-72");

        #we are now ready to collect the results of the dSGP4 propagation for the same tsince values
        tsinces=torch.tensor([0.,360.,720.,1080.,1440.])
        results_dsgp4=[]
        for tle in tles:
            results_dsgp4.append(dsgp4.propagate(tle, tsinces).numpy().reshape(5,6))    
        results_dsgp4=np.stack(results_dsgp4) 
        self.assertTrue(np.allclose(results_dsgp4,space_track_results[:,:,1:],atol=1e-6))




file="""
0 COSMOS 2251 DEB
1 34427U 93036RU  22068.94647328  .00008100  00000-0  11455-2 0  9999
2 34427  74.0145 306.8269 0033346  13.0723 347.1308 14.76870515693886
0 COSMOS 2251 DEB
1 34428U 93036RV  22068.90158861  .00002627  00000-0  63561-3 0  9999
2 34428  74.0386 139.1157 0038434 196.7068 279.9147 14.53118052686950
0 COSMOS 2251 DEB
1 34429U 93036RW  22068.93069979  .00002324  00000-0  48175-3 0  9992
2 34429  74.0120 358.8898 0044194 129.5211 303.2208 14.60189990691304
0 COSMOS 2251 DEB
1 34430U 93036RX  22068.81563909  .00000483  00000-0  17604-3 0  9990
2 34430  74.0532 306.2833 0065617 358.1277   1.9618 14.34597585682237
0 COSMOS 2251 DEB
1 34431U 93036RY  22068.56203921  .00000581  00000-0  26231-3 0  9993
2 34431  74.0773 128.6403 0072810  98.6936 262.2466 14.22838740675632
0 COSMOS 2251 DEB
1 34432U 93036RZ  22068.88256954  .00000495  00000-0  31471-3 0  9999
2 34432  73.8343 350.6259 0194977 316.7149 104.7023 13.97612468664598
0 COSMOS 2251 DEB
1 34433U 93036SA  22068.32456284  .00001354  00000-0  78369-3 0  9999
2 34433  74.0206  43.9986 0175048 133.2292 228.3613 14.02398768666229
0 COSMOS 2251 DEB
1 34434U 93036SB  22069.18354189  .00002732  00000-0  20462-2 0  9997
2 34434  74.0680 145.5195 0335997 236.1481 152.1039 13.71534520650560
0 COSMOS 2251 DEB
1 34439U 93036SG  22069.19570012  .00007264  00000-0  12034-2 0  9999
2 34439  74.0612 359.9013 0119628  27.3462 333.3934 14.65738840693165
0 COSMOS 2251 DEB
1 34442U 93036SK  22068.92772226  .00002499  00000-0  43362-3 0  9998
2 34442  74.0170 264.7116 0070210  17.5471  98.7860 14.67183128695256
0 COSMOS 2251 DEB
1 34445U 93036SN  22068.73338576  .00009736  00000-0  88849-3 0  9996
2 34445  74.0124  41.2799 0085291 113.2816 310.9216 14.92389499701434
0 COSMOS 2251 DEB
1 34446U 93036SP  22068.81213567  .00000225  00000-0  68441-4 0  9999
2 34446  74.0404 108.2372 0054093 229.4135 130.2310 14.47032726688077
0 COSMOS 2251 DEB
1 34447U 93036SQ  22068.71718351  .00027570  00000-0  27166-2 0  9999
2 34447  74.0222 176.1774 0021532 277.1236 193.9740 14.91757487697531
0 COSMOS 2251 DEB
1 34449U 93036SS  22069.15291990  .00003563  00000-0  93156-3 0  9994
2 34449  74.0534 185.4595 0067494 230.9988 198.3222 14.48394442685981
0 COSMOS 2251 DEB
1 34451U 93036SU  22068.79243827  .00003868  00000-0  10233-2 0  9994
2 34451  74.0627 200.4329 0067663 231.7496 127.7563 14.47791432685399
0 COSMOS 2251 DEB
1 34452U 93036SV  22068.79425045  .00001866  00000-0  22273-3 0  9993
2 34452  74.0450  58.8413 0116185 144.3611 286.0955 14.80282157701765
0 COSMOS 2251 DEB
1 34453U 93036SW  22068.88937838  .00002522  00000-0  63547-3 0  9991
2 34453  74.1682 233.2845 0047826 119.5810   3.3000 14.50959767686490
0 COSMOS 2251 DEB
1 34454U 93036SX  22068.91971155  .00000319  00000-0  11812-3 0  9996
2 34454  74.0583 280.7094 0037596 327.9100  31.9764 14.35844873683320
0 COSMOS 2251 DEB
1 34456U 93036SZ  22068.85349922  .00005488  00000-0  13805-2 0  9997
2 34456  74.0462 226.5559 0056524 251.9784 107.5213 14.50477917685046
0 COSMOS 2251 DEB
1 34457U 93036TA  22068.86184197  .00003162  00000-0  55362-3 0  9992
2 34457  74.0015 304.7929 0101048  52.3409  79.2774 14.64945713693281
0 COSMOS 2251 DEB
1 34458U 93036TB  22068.41282814  .00004795  00000-0  11759-2 0  9993
2 34458  74.0292 153.1192 0016214 199.0657 287.1028 14.52604984687106
0 COSMOS 2251 DEB
1 34459U 93036TC  22069.08344769  .00000282  00000-0  18354-3 0  9992
2 34459  74.0784 101.5573 0196934 142.0480 219.4723 13.97520124665133
0 COSMOS 2251 DEB
1 34464U 93036TH  22068.92676281  .00002223  00000-0  40696-3 0  9998
2 34464  74.0215 280.4680 0053730  59.9090  56.2251 14.65541250694777
0 COSMOS 2251 DEB
1 34465U 93036TJ  22068.68041544  .00002908  00000-0  64829-3 0  9994
2 34465  74.0759 152.8527 0089581 164.8833 195.5040 14.54849846687367
0 COSMOS 2251 DEB
1 34468U 93036TM  22068.66449296  .00000505  00000-0  15296-3 0  9996
2 34468  74.0161 160.4309 0082400 276.6917  82.4866 14.43076958686147
0 COSMOS 2251 DEB
1 34469U 93036TN  22068.86934120  .00007221  00000-0  12494-2 0  9992
2 34469  74.0174 358.6480 0010578 107.0254 253.2076 14.68476996692029
0 COSMOS 2251 DEB
1 34472U 93036TR  22069.05485062  .00001254  00000-0  10568-2 0  9994
2 34472  74.0854 185.8782 0344216 270.6029  85.5662 13.64889187650003
0 COSMOS 2251 DEB
1 34473U 93036TS  22068.77072253  .00004133  00000-0  10903-2 0  9999
2 34473  74.0323 191.9444 0023667 209.7653 274.0908 14.49104113686503
0 COSMOS 2251 DEB
1 34474U 93036TT  22069.06593345  .00001669  00000-0  10117-2 0  9991
2 34474  74.0826 178.2287 0229473 223.6924 311.6293 13.94903863663209
0 COSMOS 2251 DEB
1 34476U 93036TV  22068.63364977  .00001641  00000-0  12974-2 0  9995
2 34476  74.0120 200.3933 0366042 353.1964   6.4250 13.65164651647891
0 COSMOS 2251 DEB
1 34477U 93036TW  22069.22728087  .00001096  00000-0  30407-3 0  9994
2 34477  73.9823 137.9328 0092528 285.4009 128.2922 14.45328075687473
0 COSMOS 2251 DEB
1 34478U 93036TX  22068.38240896  .00002447  00000-0  57864-3 0  9995
2 34478  74.0630 146.6712 0096906 357.3956 122.6783 14.51844358687860
0 COSMOS 2251 DEB
1 34481U 93036UA  22068.93573963  .00000769  00000-0  32254-3 0  9995
2 34481  73.9822  34.7473 0076080 159.1313 262.1452 14.26043776678288
0 COSMOS 2251 DEB
1 34482U 93036UB  22068.94395123  .00001626  00000-0  50637-3 0  9995
2 34482  74.0495 285.7402 0073443 330.2917 141.2161 14.40195520682649
0 IRIDIUM 33 DEB
1 34486U 97051GB  22068.70192972  .00002799  00000-0  67048-3 0  9997
2 34486  86.4080 256.1055 0019315 167.8443 327.5296 14.53105816688403
0 IRIDIUM 33 DEB
1 34487U 97051GC  22068.94053602  .00000335  00000-0  99357-4 0  9993
2 34487  86.3423 241.7417 0031623 345.4457 146.4254 14.40142410684585
0 IRIDIUM 33 DEB
1 34488U 97051GD  22068.64888950  .00003454  00000-0  76260-3 0  9995
2 34488  86.3495 209.5295 0026344 115.9819 244.4107 14.56862511689744
0 IRIDIUM 33 DEB
1 34489U 97051GE  22068.70986860  .00005650  00000-0  16117-2 0  9996
2 34489  86.4090 318.3848 0074197 128.7816 232.0044 14.43442216679898
0 IRIDIUM 33 DEB
1 34490U 97051GF  22069.05151916  .00096914  00000-0  60495-2 0  9994
2 34490  86.3078 166.2021 0026908 104.4790  16.6375 15.08757669694991
0 IRIDIUM 33 DEB
1 34492U 97051GH  22068.21456637  .00003667  00000-0  20620-2 0  9990
2 34492  86.4992 148.4187 0300676 196.1027 197.7228 13.89878655657771
0 IRIDIUM 33 DEB
1 34493U 97051GJ  22068.71924692  .00015517  00000-0  23522-2 0  9998
2 34493  86.2631 163.3558 0057545 248.7354 110.7712 14.72914243689392
0 IRIDIUM 33 DEB
1 34496U 97051GM  22068.11986750  .00008875  00000-0  22227-2 0  9998
2 34496  86.3807 313.2362 0079654 214.8210 144.7768 14.49411405507528
0 IRIDIUM 33 DEB
1 34497U 97051GN  22069.07755535  .00000714  00000-0  30910-3 0  9993
2 34497  86.2760 274.9045 0091436  57.6087  72.1578 14.20664684674744
0 IRIDIUM 33 DEB
1 34503U 97051GU  22068.79645465  .00004812  00000-0  97823-3 0  9998
2 34503  86.3205 182.9085 0013065 242.5442 117.4440 14.60894160691301
0 IRIDIUM 33 DEB
1 34508U 97051GZ  22069.15122779  .00001244  00000-0  50236-3 0  9995
2 34508  86.3768 317.1334 0077286 309.0603  50.3727 14.25578501676677
0 IRIDIUM 33 DEB
1 34511U 97051HC  22068.81147505  .00005535  00000-0  13651-2 0  9991
2 34511  86.3845 264.6079 0030984 200.9344 169.6160 14.51723898685578
0 IRIDIUM 33 DEB
1 34517U 97051HJ  22068.85833177  .00022186  00000-0  23199-2 0  9997
2 34517  86.3620 190.4031 0002835  70.8380  53.2461 14.89548326694032
0 IRIDIUM 33 DEB
1 34518U 97051HK  22068.87864802  .00040197  00000-0  38222-2 0  9994
2 34518  86.3834 211.8107 0021328 221.9151 138.0446 14.93103181693502
0 IRIDIUM 33 DEB
1 34520U 97051HM  22068.88279414  .00067561  00000-0  46845-2 0  9991
2 34520  86.3574 191.0984 0004408 115.1070 245.0623 15.05298769694229
0 IRIDIUM 33 DEB
1 34521U 97051HN  22068.82096123  .00001811  00000-0  52005-3 0  9992
2 34521  86.3823 263.4130 0006431 189.2578 170.8506 14.44393233685383
0 IRIDIUM 33 DEB
1 34522U 97051HP  22068.93100740  .00024441  00000-0  40753-2 0  9996
2 34522  86.3626 249.0936 0048383 205.8879 303.0979 14.69081619686407
0 IRIDIUM 33 DEB
1 34524U 97051HR  22069.18279791  .00007169  00000-0  16785-2 0  9995
2 34524  86.3435 235.2795 0039177 195.3316 164.6708 14.54007453686529
0 IRIDIUM 33 DEB
1 34525U 97051HS  22068.83195944  .00000608  00000-0  19645-3 0  9994
2 34525  86.3704 269.8024 0002882 299.3778  72.1459 14.37508703683198
0 IRIDIUM 33 DEB
1 34526U 97051HT  22068.91198510  .00026319  00000-0  30963-2 0  9998
2 34526  86.3397 196.8887 0017449 225.1990 134.7816 14.84614878691814
0 IRIDIUM 33 DEB
1 34529U 97051HW  22068.77403985  .00000428  00000-0  13464-3 0  9991
2 34529  86.3611 259.3456 0007967  48.6875 311.5008 14.38083903683803
0 IRIDIUM 33 DEB
1 34532U 97051HZ  22068.93917639  .00029831  00000-0  40418-2 0  9996
2 34532  86.3340 225.2890 0034538  76.4842 284.0222 14.78381229687449
0 IRIDIUM 33 DEB
1 34535U 97051JC  22068.72880154  .00021552  00000-0  22690-2 0  9991
2 34535  86.3316 164.9437 0023416  87.7299 272.6608 14.89072483695659
0 IRIDIUM 33 DEB
1 34538U 97051JF  22068.71389750  .00001304  00000-0  47311-3 0  9990
2 34538  86.2789 240.6891 0045457 130.9806   1.3945 14.32050243680322
0 IRIDIUM 33 DEB
1 34540U 97051JH  22069.09230438  .00001930  00000-0  45493-3 0  9997
2 34540  86.3285 200.0620 0028056 167.6445   5.2521 14.53569147690009
0 COSMOS 2251 DEB
1 34542U 93036UC  22069.12781667  .00004135  00000-0  70743-3 0  9993
2 34542  74.0223 311.9362 0024129  80.2712 280.1181 14.68987571692184
0 COSMOS 2251 DEB
1 34544U 93036UE  22068.94281865  .00000383  00000-0  37993-3 0  9990
2 34544  73.9154 262.0212 0393317 184.2596 297.2850 13.51829391643005
0 COSMOS 2251 DEB
1 34545U 93036UF  22068.39378412  .00007551  00000-0  14486-2 0  9990
2 34545  74.0362  56.3228 0010568 112.5254 247.7033 14.63848827690137
0 COSMOS 2251 DEB
1 34546U 93036UG  22069.02028708  .00000734  00000-0  25473-3 0  9997
2 34546  74.0328 270.3327 0024582 320.8666 192.7630 14.37197749682875
0 COSMOS 2251 DEB
1 34547U 93036UH  22068.73290787  .00002426  00000-0  67758-3 0  9995
2 34547  74.0407 186.0122 0036889 249.0105 110.7105 14.46361434685248
0 COSMOS 2251 DEB
1 34550U 93036UL  22068.48847750  .00005213  00000-0  16483-2 0  9994
2 34550  73.9920 106.8624 0177754 242.5189 115.7794 14.31752709674389
0 COSMOS 2251 DEB
1 34552U 93036UN  22069.14806022  .00004264  00000-0  10926-2 0  9998
2 34552  74.0336 195.5332 0031631 222.1475 202.7766 14.50342128684399
0 COSMOS 2251 DEB
1 34553U 93036UP  22069.20770096  .00001545  00000-0  48719-3 0  9998
2 34553  74.0298 242.2505 0026659 293.3152 241.7380 14.40957065682782
0 COSMOS 2251 DEB
1 34554U 93036UQ  22068.68224066  .00000410  00000-0  15516-3 0  9990
2 34554  74.0390 300.0901 0024846 340.2587  75.5590 14.34163705680359
0 COSMOS 2251 DEB
1 34555U 93036UR  22069.20221691  .00004198  00000-0  97755-3 0  9993
2 34555  74.0268 137.2536 0021907 183.9846 241.7047 14.55010683686251
0 COSMOS 2251 DEB
1 34556U 93036US  22068.81781138  .00002697  00000-0  77989-3 0  9996
2 34556  74.0310 214.3189 0022536 252.7941 224.2692 14.44887741683244
0 COSMOS 2251 DEB
1 34560U 93036UW  22068.57206785  .00002190  00000-0  64784-3 0  9991
2 34560  74.0278 222.2738 0029016 276.5018 155.9951 14.43797393682765
0 COSMOS 2251 DEB
1 34561U 93036UX  22068.81866329  .00002883  00000-0  63805-3 0  9997
2 34561  74.0322  89.5651 0038908 166.5515 193.6696 14.57211065687340
0 COSMOS 2251 DEB
1 34563U 93036UZ  22068.91980734  .00000721  00000-0  23814-3 0  9992
2 34563  74.0392 232.3082 0015247 303.3138 183.2627 14.39779493683864
0 COSMOS 2251 DEB
1 34564U 93036VA  22069.19372862  .00000400  00000-0  14311-3 0  9997
2 34564  74.0322 250.9380 0020030 331.0174 151.3479 14.37205198683156
0 COSMOS 2251 DEB
1 34567U 93036VD  22068.53031995  .00000275  00000-0  13201-3 0  9993
2 34567  74.0407 105.3088 0075535 150.0956 329.2084 14.21959620675796
0 COSMOS 2251 DEB
1 34569U 93036VF  22068.95182238  .00000620  00000-0  27037-3 0  9994
2 34569  73.9945 147.7061 0171535 287.6536 248.8497 14.18538313674513
0 COSMOS 2251 DEB
1 34570U 93036VG  22068.47157709  .00000492  00000-0  32097-3 0  9997
2 34570  73.9598  78.8025 0201588 249.6357 264.3030 13.95626211664035
0 COSMOS 2251 DEB
1 34571U 93036VH  22067.56494129  .00002866  00000-0  12100-2 0  9996
2 34571  74.0520 174.2016 0101176 188.3648 171.5817 14.23322933676994
0 COSMOS 2251 DEB
1 34572U 93036VJ  22068.24483215  .00001990  00000-0  80964-3 0  9998
2 34572  74.0185 118.5245 0096517 160.3823 267.8551 14.25653884676611
0 COSMOS 2251 DEB
1 34574U 93036VL  22068.59127556  .00001609  00000-0  39375-3 0  9998
2 34574  74.0759 133.0550 0095628 158.9867 324.0525 14.50605175687532
0 COSMOS 2251 DEB
1 34576U 93036VN  22068.74071590  .00001339  00000-0  40122-3 0  9995
2 34576  74.0404 191.8549 0010258 293.6624 189.3232 14.43753808685720
0 COSMOS 2251 DEB
1 34577U 93036VP  22068.94317596  .00002753  00000-0  56316-3 0  9990
2 34577  74.0392   4.0563 0032171 131.1912 229.2037 14.60988208691606
0 COSMOS 2251 DEB
1 34579U 93036VR  22068.92726392  .00000618  00000-0  20689-3 0  9990
2 34579  74.0542 246.6393 0030227 299.6184 178.5935 14.39172196683758
0 COSMOS 2251 DEB
1 34582U 93036VU  22068.94433895  .00000723  00000-0  25578-3 0  9998
2 34582  74.0213 288.2871 0055074  10.5073 349.7216 14.35548910681898
0 COSMOS 2251 DEB
1 34585U 93036VX  22069.12693092  .00006742  00000-0  10695-2 0  9999
2 34585  74.0265 319.1964 0013296  85.4540 274.8150 14.72289235692389
0 COSMOS 2251 DEB
1 34588U 93036WA  22068.96185243  .00000495  00000-0  17801-3 0  9993
2 34588  74.0438 280.1256 0020513 313.6191 152.3758 14.36328209681931
0 COSMOS 2251 DEB
1 34589U 93036WB  22068.49627174  .00000424  00000-0  31782-3 0  9992
2 34589  73.9241 230.2365 0258066  83.6404 333.3486 13.82672155656973
0 COSMOS 2251 DEB
1 34590U 93036WC  22068.95802757  .00000213  00000-0  57730-4 0  9995
2 34590  74.0449  35.1009 0071997 153.2050 269.3897 14.52222611691633
0 COSMOS 2251 DEB
1 34609U 93036WE  22068.78185174  .00000613  00000-0  29573-3 0  9991
2 34609  73.7342  80.4141 0158112 159.5799  13.4031 14.14491694671628
0 COSMOS 2251 DEB
1 34610U 93036WF  22067.60352825  .00001705  00000-0  10416-2 0  9990
2 34610  74.0258 129.8069 0207283 213.2644 145.5320 13.96626537663004
0 COSMOS 2251 DEB
1 34611U 93036WG  22069.12250983  .00001660  00000-0  45802-3 0  9990
2 34611  74.0641 180.5236 0037461 223.4734 201.6638 14.47193983684722
0 COSMOS 2251 DEB
1 34613U 93036WJ  22068.72863866  .00001082  00000-0  53831-3 0  9996
2 34613  73.9681 223.5130 0129564 354.9799  16.8811 14.13981619669720
0 COSMOS 2251 DEB
1 34614U 93036WK  22068.63183764  .00004007  00000-0  24848-2 0  9994
2 34614  73.8895 226.7217 0273044 126.6005   3.7416 13.88618176655351
0 COSMOS 2251 DEB
1 34615U 93036WL  22069.18146051  .00000335  00000-0  33390-3 0  9997
2 34615  74.0062 331.1428 0409982 159.0926 327.4818 13.49468644642707
0 COSMOS 2251 DEB
1 34616U 93036WM  22068.66873903  .00079618  00000-0  50636-2 0  9994
2 34616  74.0152 264.4423 0009130 189.7533 287.2375 15.08363272693295
0 COSMOS 2251 DEB
1 34619U 93036WQ  22068.93716744  .00000494  00000-0  10379-3 0  9993
2 34619  74.0476 306.5313 0101351  34.0169 326.7438 14.59259854692280
0 COSMOS 2251 DEB
1 34620U 93036WR  22068.72189948  .00003328  00000-0  14138-2 0  9997
2 34620  74.0686 203.0250 0123920 221.3081 137.8643 14.21617239671578
0 COSMOS 2251 DEB
1 34622U 93036WT  22068.83487469  .00000686  00000-0  34723-3 0  9997
2 34622  74.0424 234.4368 0119529 285.2354 141.9703 14.14417339669908
0 COSMOS 2251 DEB
1 34626U 93036WX  22069.12334337  .00001420  00000-0  33666-3 0  9991
2 34626  74.0573  68.4773 0038853 158.9112 271.0263 14.54489517690888
0 COSMOS 2251 DEB
1 34629U 93036XA  22068.48260093  .00014277  00000-0  11303-2 0  9996
2 34629  73.7784 227.2057 0054198 261.9499  97.5534 14.99449156702738
0 COSMOS 2251 DEB
1 34634U 93036XF  22068.86553663  .00000502  00000-0  13689-3 0  9998
2 34634  74.0521 100.2999 0051675 197.9671 217.4335 14.49249934688427
0 COSMOS 2251 DEB
1 34635U 93036XG  22068.70033788  .00001820  00000-0  48750-3 0  9994
2 34635  74.0537 153.8664 0046265 215.2520 267.6548 14.48319810532212
0 COSMOS 2251 DEB
1 34637U 93036XJ  22068.72107742  .00001738  00000-0  47830-3 0  9996
2 34637  74.0547 166.9932 0066667 226.8243 251.2590 14.46396447685850
0 COSMOS 2251 DEB
1 34638U 93036XK  22068.91656580  .00003294  00000-0  73037-3 0  9994
2 34638  74.0435 106.0047 0008556 157.1245 261.9065 14.57500207687535
0 COSMOS 2251 DEB
1 34639U 93036XL  22068.68400142  .00004423  00000-0  10283-2 0  9998
2 34639  74.0334 147.5146 0043510 201.4877 286.4544 14.54631374685390
0 COSMOS 2251 DEB
1 34641U 93036XN  22068.91549830  .00001103  00000-0  36055-3 0  9993
2 34641  73.8436 149.8823 0020840 334.5142 146.7654 14.39691670684166
0 COSMOS 2251 DEB
1 34642U 93036XP  22069.01466616  .00000998  00000-0  66107-3 0  9998
2 34642  74.0075 181.2175 0227764 302.2655 174.4725 13.90985389661918
0 IRIDIUM 33 DEB
1 34643U 97051JU  22068.68681555  .00006654  00000-0  14774-2 0  9990
2 34643  86.3796 251.6356 0035333 160.5044 199.7525 14.56546284683433
0 IRIDIUM 33 DEB
1 34648U 97051JZ  22069.10136961  .00002815  00000-0  18997-2 0  9995
2 34648  86.4076 101.5871 0230215 199.1784 160.0616 13.88148366658369
0 IRIDIUM 33 DEB
1 34651U 97051KC  22068.20655725  .00001931  00000-0  10436-2 0  9991
2 34651  86.2956 352.1986 0188805 205.0191 154.1750 14.03506581664900
0 IRIDIUM 33 DEB
1 34652U 97051KD  22068.71059411  .00001745  00000-0  46623-3 0  9999
2 34652  86.3381 222.4216 0036042 230.3439 140.9446 14.47457461686011
0 NAVSTAR 63 (USA 203)
1 34661U 09014A   22068.49766742  .00000025  00000-0  00000-0 0  9994
2 34661  55.1247 273.3242 0121622  55.4323 306.7902  2.00559615 94985
0 SL-12 DEB
1 34667U 91025L   22069.09518701  .00000568  00000-0  86705-2 0  9990
2 34667  65.8889 280.5501 5458247  92.4487 357.8727  4.35288851205462
0 COSMOS 2251 DEB
1 34671U 93036XR  22068.82793444  .00001830  00000-0  53720-3 0  9993
2 34671  74.0351 206.7640 0021166 263.1629 222.0589 14.44403303683442
0 COSMOS 2251 DEB
1 34672U 93036XS  22068.41156390  .00000300  00000-0  11768-3 0  9999
2 34672  74.0352 303.6683 0025348 351.0783 182.6152 14.33502764454679
0 COSMOS 2251 DEB
1 34674U 93036XU  22068.76822552  .00003298  00000-0  58970-3 0  9994
2 34674  73.9845 332.0633 0111666 257.4278 101.4382 14.63317013690540
0 COSMOS 2251 DEB
1 34675U 93036XV  22068.85898012  .00003276  00000-0  76511-3 0  9996
2 34675  74.0348 137.3326 0032168 196.8713 163.1383 14.54798835686024
0 COSMOS 2251 DEB
1 34678U 93036XY  22068.85041662  .00000025  00000-0  19743-4 0  9994
2 34678  74.0395 342.6753 0025400   2.9075  58.8849 14.29731025677424
0 COSMOS 2251 DEB
1 34679U 93036XZ  22068.83834898  .00005961  00000-0  93844-3 0  9991
2 34679  74.0089 277.5975 0042065  22.4234 337.8764 14.72112204693342
0 COSMOS 2251 DEB
1 34680U 93036YA  22068.46794456  .00001710  00000-0  49899-3 0  9996
2 34680  74.0494 193.3378 0011924 263.2830 168.3865 14.44826463683945
0 COSMOS 2251 DEB
1 34681U 93036YB  22068.71717113  .00002091  00000-0  46957-3 0  9995
2 34681  74.0746  54.7516 0040284 121.5808  51.1250 14.56677786691279
0 COSMOS 2251 DEB
1 34682U 93036YC  22068.91068493  .00001075  00000-0  30263-3 0  9992
2 34682  74.0441 149.1970 0024657 260.9363 222.5317 14.46901496687499
0 COSMOS 2251 DEB
1 34683U 93036YD  22068.72708171  .00001739  00000-0  48845-3 0  9998
2 34683  74.0362 167.4910 0028868 257.3547 222.0571 14.46457811683851
0 COSMOS 2251 DEB
1 34685U 93036YF  22069.15636324  .00001324  00000-0  40077-3 0  9997
2 34685  74.0353 204.2239 0019721 280.8056 253.8297 14.43205135686892
0 COSMOS 2251 DEB
1 34687U 93036YH  22069.09795462  .00009616  00000-0  67555-3 0  9993
2 34687  73.9226 211.9798 0083273  41.8298  74.6081 15.02397782707016
0 COSMOS 2251 DEB
1 34689U 93036YK  22068.39672330  .00002967  00000-0  20466-2 0  9996
2 34689  74.1011  41.0080 0284977  81.2570  69.9302 13.81824768652313
0 IRIDIUM 33 DEB
1 34690U 97051KM  22068.83364907  .00001478  00000-0  43019-3 0  9992
2 34690  86.3819 266.9213 0008297 294.2361 126.7797 14.43629843682055
0 IRIDIUM 33 DEB
1 34693U 97051KQ  22068.90768167  .00003071  00000-0  68981-3 0  9995
2 34693  86.3422 208.0805 0027161 123.3364   0.2315 14.56008043686657
0 IRIDIUM 33 DEB
1 34696U 97051KT  22069.01427476  .00002379  00000-0  10771-2 0  9999
2 34696  86.2042 272.0186 0139501 313.0537 200.2712 14.16615764670478
0 IRIDIUM 33 DEB
1 34702U 97051KZ  22068.50074806  .00003193  00000-0  62095-3 0  9997
2 34702  86.3131 172.3314 0038928 359.9862   0.1341 14.62357707691099
0 IRIDIUM 33 DEB
1 34705U 97051LC  22068.85690611  .00013199  00000-0  16888-2 0  9991
2 34705  86.4603 285.1272 0073387   0.1236 359.9989 14.79345575689044
0 IRIDIUM 33 DEB
1 34706U 97051LD  22068.93861435  .00037695  00000-0  23337-2 0  9999
2 34706  86.3021 126.0688 0018169  51.3179  68.4484 15.09494923698029
0 EUTE 10A (EUTE W2A)
1 34710U 09016A   22068.90577198  .00000065  00000-0  00000-0 0  9996
2 34710   0.0616   1.6065 0005540 350.7126 151.2827  1.00272168 25613
0 BREEZE-M R/B
1 34711U 09016B   22068.65983842 -.00000007  00000-0  00000-0 0  9993
2 34711  20.5652 253.1520 5936734 183.9473 167.5621  2.20367276112140
0 BREEZE-M DEB (TANK)
1 34712U 09016C   22068.68120304  .00003542  00000-0  51437-3 0  9992
2 34712  46.0467 243.9004 5042943 233.3680  69.0197  5.55861109255565
0 WGS F2 (USA 204)
1 34713U 09017A   22069.13046942  .00000065  00000-0  00000+0 0  9993
2 34713   0.0173 160.5373 0001922  79.0554  32.6592  1.00270811 47299
0 ATLAS 5 CENTAUR R/B
1 34714U 09017B   22066.24526347  .00001011  00000-0  35466-2 0  9994
2 34714  16.8402 252.9671 8246482 246.5776  13.9459  1.16266407 54288
0 COSMOS 2251 DEB
1 34721U 93036YP  22069.05735669  .00000876  00000-0  59488-3 0  9996
2 34721  73.9246 170.2277 0238990  22.0153 151.6200 13.88659436660347
0 COSMOS 2251 DEB
1 34722U 93036YQ  22068.80660747  .00000515  00000-0  28055-3 0  9999
2 34722  73.8871 252.3946 0178278  86.0938 287.8838 14.06747396668946
0 COSMOS 2251 DEB
1 34723U 93036YR  22067.64909525  .12226285  19405-5  60915-2 0  9992
2 34723  73.9258 293.0276 0008565 260.5674  99.4607 16.20633215700335
0 COSMOS 2251 DEB
1 34724U 93036YS  22068.95171465  .00009396  00000-0  44780-2 0  9999
2 34724  74.0463 170.8083 0235314 247.2857 110.3254 14.05916928659456
0 COSMOS 2251 DEB
1 34725U 93036YT  22068.47891361  .00017653  00000-0  28047-2 0  9994
2 34725  74.2020 197.1756 0038030 342.3895  17.5956 14.71586483686697
0 COSMOS 2251 DEB
1 34726U 93036YU  22068.76244980  .00000861  00000-0  22039-3 0  9995
2 34726  74.0287  66.8109 0051070 200.3154 159.5977 14.51169034685242
0 COSMOS 2251 DEB
1 34728U 93036YW  22068.88428617  .00001155  00000-0  26121-3 0  9994
2 34728  74.0252   5.3476 0054228 149.2994 211.1362 14.56491785687197
0 COSMOS 2251 DEB
1 34729U 93036YX  22069.15419014  .00001673  00000-0  42481-3 0  9999
2 34729  74.0282 112.0535 0030001 218.5248 208.0466 14.51309990684208
0 COSMOS 2251 DEB
1 34732U 93036ZA  22068.45347061  .00000237  00000-0  74752-4 0  9999
2 34732  74.0447 136.1794 0042142 270.2175 101.0719 14.45188364682789
0 COSMOS 2251 DEB
1 34735U 93036ZD  22068.29627898  .00009433  00000-0  12823-2 0  9999
2 34735  74.0294 282.7872 0047738 336.6650  23.2357 14.78105639692603
0 COSMOS 2251 DEB
1 34736U 93036ZE  22068.79062323  .00003254  00000-0  86845-3 0  9992
2 34736  74.0360 182.9041 0017208 226.1241 251.9473 14.48715830682587
0 COSMOS 2251 DEB
1 34737U 93036ZF  22068.83700856  .00009138  00000-0  16175-2 0  9993
2 34737  74.0362  85.5052 0022213   0.9773  53.0536 14.67285749686026
0 COSMOS 2251 DEB
1 34738U 93036ZG  22068.77818821  .00000488  00000-0  15521-3 0  9993
2 34738  74.0189 180.3641 0050867 297.8068 189.0747 14.41834665680300
0 COSMOS 2251 DEB
1 34739U 93036ZH  22068.34657073  .00010594  00000-0  13328-2 0  9998
2 34739  74.0264 182.5803 0055026 262.4966 159.5514 14.81060712696338
0 COSMOS 2251 DEB
1 34741U 93036ZK  22069.14364918  .00046838  00000-0  20549-2 0  9996
2 34741  73.9931 346.1735 0064179  66.4127 294.3792 15.19854917702538
0 COSMOS 2251 DEB
1 34744U 93036ZN  22068.82408244  .00005846  00000-0  89039-3 0  9998
2 34744  74.0081 265.8069 0036861  14.7410 357.0262 14.73695161692141
0 COSMOS 2251 DEB
1 34745U 93036ZP  22069.16749933  .00044130  00000-0  26656-2 0  9994
2 34745  74.0130  89.1565 0012838 237.2251 122.7705 15.10411539698892
0 COSMOS 2251 DEB
1 34746U 93036ZQ  22068.87299352  .00007025  00000-0  12229-2 0  9990
2 34746  74.0281 359.6835 0019520  79.2341 336.2123 14.68125772691425
0 COSMOS 2251 DEB
1 34747U 93036ZR  22068.83221079  .00015800  00000-0  13067-2 0  9999
2 34747  73.9864 355.9292 0044633 144.7464 215.6688 14.98129691701788
0 COSMOS 2251 DEB
1 34750U 93036ZU  22068.69062684  .00001609  00000-0  47293-3 0  9993
2 34750  74.1154 252.4121 0053097 217.5174 268.5978 14.43757078683076
0 COSMOS 2251 DEB
1 34751U 93036ZV  22068.85636538  .00005175  00000-0  13522-2 0  9996
2 34751  74.0067 215.2526 0035843 235.1604 248.2185 14.49267877683702
0 COSMOS 2251 DEB
1 34752U 93036ZW  22069.04619078  .00001738  00000-0  27856-3 0  9996
2 34752  74.0239 192.7702 0089161 313.3233  46.0528 14.69940591695204
0 COSMOS 2251 DEB
1 34754U 93036ZY  22068.90887756  .00000301  00000-0  10871-3 0  9996
2 34754  74.0301 241.9607 0030919 333.0041 149.7006 14.37578019683745
0 COSMOS 2251 DEB
1 34755U 93036ZZ  22068.96073263  .00000153  00000-0  16973-3 0  9997
2 34755  73.9295 329.8585 0420789 244.4607 124.2380 13.46911975640060
0 COSMOS 2251 DEB
1 34756U 93036AAA 22068.69017199  .00000626  00000-0  54737-3 0  9999
2 34756  74.0048 191.9284 0361339 354.7345   5.0006 13.61679417646855
0 COSMOS 2251 DEB
1 34758U 93036AAC 22068.80215110  .00001416  00000-0  12909-2 0  9999
2 34758  74.1098 341.4888 0398159  61.1150 302.9298 13.53423274642934
0 COSMOS 2251 DEB
1 34759U 93036AAD 22068.85701524  .00015822  00000-0  22165-2 0  9995
2 34759  74.0166   4.0855 0010827   5.3999 354.7291 14.77482314693559
0 COSMOS 2251 DEB
1 34760U 93036AAE 22069.19752550  .00000866  00000-0  43707-3 0  9998
2 34760  74.0644 223.2292 0102496 251.3660 284.0339 14.15081754672644
0 COSMOS 2251 DEB
1 34761U 93036AAF 22068.52064119  .00000971  00000-0  37594-3 0  9996
2 34761  74.0333 358.0392 0041935   5.3956 354.7637 14.30896942680097
0 COSMOS 2251 DEB
1 34763U 93036AAH 22068.78065034  .00008832  00000-0  18512-2 0  9995
2 34763  74.1730 313.6296 0064750 119.4211 308.0251 14.58493953683870
0 IRIDIUM 33 DEB
1 34764U 97051LH  22068.57543325  .00054250  00000-0  39325-2 0  9998
2 34764  86.3538 178.2573 0005954 260.8955  99.1605 15.03689251694515
0 IRIDIUM 33 DEB
1 34765U 97051LJ  22068.78051463  .00000657  00000-0  22352-3 0  9994
2 34765  86.3366 258.8957 0018995  46.0804 323.8526 14.35014857678780
0 IRIDIUM 33 DEB
1 34773U 97051LS  22068.71126542  .00004841  00000-0  10962-2 0  9999
2 34773  86.4072 264.8974 0017105  72.0967 288.2104 14.55898486688104
0 IRIDIUM 33 DEB
1 34774U 97051LT  22069.05934824  .00004012  00000-0  79091-3 0  9997
2 34774  86.4186 259.0488 0021525 326.6878 158.1491 14.62151183690726
0 IRIDIUM 33 DEB
1 34775U 97051LU  22068.81064666  .00002627  00000-0  74056-3 0  9995
2 34775  86.3753 269.4146 0026287 281.8282  89.7599 14.45272552682299
0 COMPASS G2
1 34779U 09018A   22068.02701842 -.00000151  00000-0  00000-0 0  9998
2 34779   8.2706  50.6554 0319604 249.7419 106.8765  0.94516854 47397
0 CZ-3C R/B
1 34780U 09018B   22069.16415195  .00005617  00000-0  24655-2 0  9994
2 34780  20.7006 321.7320 7131875  39.9300 354.9073  2.45578427112046
0 COSMOS 2251 DEB
1 34784U 93036AAM 22068.36638343  .00000887  00000-0  22329-3 0  9995
2 34784  74.0090  52.3058 0060054 197.0015 162.9136 14.51648502685406
0 COSMOS 2251 DEB
1 34787U 93036AAQ 22069.09891571  .00006428  00000-0  10002-2 0  9994
2 34787  74.0161 309.5714 0013326  73.4927 286.7708 14.73138766692625
0 COSMOS 2251 DEB
1 34788U 93036AAR 22068.94122777  .00003389  00000-0  29271-3 0  9995
2 34788  73.9903 280.8094 0134553  31.4587  33.2058 14.90884651707015
0 COSMOS 2251 DEB
1 34789U 93036AAS 22068.79813514  .00004225  00000-0  14561-2 0  9990
2 34789  74.2048 192.1639 0103795  18.1297 106.5324 14.33192302674468
0 COSMOS 2251 DEB
1 34790U 93036AAT 22068.77443678  .00000752  00000-0  35584-3 0  9990
2 34790  74.0573 182.7815 0109231 218.6289 269.6003 14.18323488670567
0 COSMOS 2251 DEB
1 34791U 93036AAU 22068.74233478  .00008295  00000-0  85591-3 0  9996
2 34791  74.0186  83.2071 0046207 213.6720 146.1524 14.89529078701119
0 COSMOS 2251 DEB
1 34792U 93036AAV 22068.52062462  .00001556  00000-0  48324-3 0  9992
2 34792  74.0244 222.0367 0016818 284.6326 138.2886 14.41827284677664
0 COSMOS 2251 DEB
1 34793U 93036AAW 22068.80776917  .00000342  00000-0  12642-3 0  9997
2 34793  74.0020 260.9066 0053385  18.5431 353.3504 14.35371548675985
0 COSMOS 2251 DEB
1 34796U 93036AAZ 22068.93184426  .00000768  00000-0  23839-3 0  9994
2 34796  74.0333 248.2516 0116421 327.7024 155.9459 14.39028736682118
0 COSMOS 2251 DEB
1 34797U 93036ABA 22068.80523530  .00001512  00000-0  62383-3 0  9999
2 34797  73.9824  95.8766 0092337 161.8617 198.5862 14.25350974677447
0 RISAT 2
1 34807U 09019A   22069.20015293  .00039560  00000-0  69476-3 0  9990
2 34807  41.2102 146.3902 0010166  13.1639  10.3869 15.49708764724344
0 BLOCK DM-SL R/B
1 34811U 09020B   22068.85335389 -.00000141  00000-0  00000-0 0  9997
2 34811   2.0707  53.4388 4737112 332.3433 217.9075  1.81487904 85390
0 COSMOS 2251 DEB
1 34813U 93036ABC 22067.29481919  .00000681  00000-0  67574-3 0  9997
2 34813  73.9353 154.4252 0489767  82.7065 282.9413 13.36780255631978
0 COSMOS 2251 DEB
1 34814U 93036ABD 22069.17118986  .00000858  00000-0  34724-3 0  9995
2 34814  73.8993 331.5394 0062383 142.3544   8.9793 14.28255681677530
0 COSMOS 2251 DEB
1 34815U 93036ABE 22068.91232080  .00001967  00000-0  40442-3 0  9995
2 34815  74.0487  10.1302 0079765  75.9212 285.0802 14.59278319684866
0 COSMOS 2251 DEB
1 34816U 93036ABF 22068.65129535  .00000759  00000-0  70667-3 0  9997
2 34816  73.8883 217.3255 0386360 167.6791 208.1205 13.54666441641547
0 COSMOS 2251 DEB
1 34818U 93036ABH 22068.66375868  .00000290  00000-0  11299-3 0  9998
2 34818  74.0358 296.5088 0025645 348.0781  74.7604 14.34015414676569
0 COSMOS 2251 DEB
1 34819U 93036ABJ 22068.81323212  .00002014  00000-0  60010-3 0  9993
2 34819  74.0258 210.7249 0024756 274.5436 213.0924 14.43557431682951
0 COSMOS 2251 DEB
1 34822U 93036ABM 22069.22762969  .00023707  00000-0  21298-2 0  9995
2 34822  74.0087 165.0218 0016564 274.5176  85.4115 14.95514971691617
0 COSMOS 2251 DEB
1 34823U 93036ABN 22067.50785894  .00001244  00000-0  29904-3 0  9992
2 34823  74.0453 161.5962 0180429 231.2404 127.2515 14.45063959684608
0 IRIDIUM 33 DEB
1 34825U 97051LZ  22068.47308345  .00006410  00000-0  14212-2 0  9999
2 34825  86.3705 241.9781 0013983  91.3791 326.5475 14.56945893684250
0 IRIDIUM 33 DEB
1 34827U 97051MB  22068.61818156  .00024454  00000-0  21723-2 0  9996
2 34827  86.3796 199.0310 0013213 235.3680 135.6571 14.95959056691620
0 IRIDIUM 33 DEB
1 34833U 97051MH  22068.83078994  .00001060  00000-0  32400-3 0  9998
2 34833  86.3930 275.9629 0022490 342.7752  28.5662 14.40859909678671
0 YAOGAN 6
1 34839U 09021A   22068.92564569  .00003565  00000-0  12986-3 0  9991
2 34839  97.0742  53.9156 0024904 294.8291  65.0360 15.28530870716048
0 COSMOS 2251 DEB
1 34845U 93036ABP 22068.57974821  .00000498  00000-0  15997-3 0  9991
2 34845  74.0284 183.6066 0023709 318.0798  41.8542 14.41875920681592
0 COSMOS 2251 DEB
1 34846U 93036ABQ 22068.63782815  .00000997  00000-0  34479-3 0  9999
2 34846  74.0339 285.6547 0030208 329.5499  30.3899 14.36804163676102
0 COSMOS 2251 DEB
1 34847U 93036ABR 22068.37367386  .00004303  00000-0  38787-3 0  9994
2 34847  74.0454 315.4655 0108832  17.3330 343.1520 14.91420236474229
0 COSMOS 2251 DEB
1 34850U 93036ABU 22068.84071782  .00010823  00000-0  17187-2 0  9996
2 34850  74.0111 349.4845 0046607  67.6146 347.9662 14.71457370687430
0 COSMOS 2251 DEB
1 34852U 93036ABW 22067.23636372  .00000142  00000-0  58415-4 0  9996
2 34852  74.0310 269.0644 0016912 347.1365  12.9354 14.35455302677880
0 COSMOS 2251 DEB
1 34853U 93036ABX 22068.73321110  .00011509  00000-0  13742-2 0  9995
2 34853  74.0689 226.9168 0041007 235.4685 124.2620 14.83688509691924
0 COSMOS 2251 DEB
1 34854U 93036ABY 22068.37127277  .00001358  00000-0  26457-3 0  9993
2 34854  74.0342 294.3970 0084028  36.2123 324.4690 14.61853450517186
0 COSMOS 2251 DEB
1 34856U 93036ACA 22069.00383345  .00001864  00000-0  33800-3 0  9998
2 34856  74.0123 270.4844 0075755  28.6563 121.6279 14.65169570688175
0 COSMOS 2251 DEB
1 34857U 93036ACB 22068.55203541  .00001505  00000-0  42244-3 0  9991
2 34857  74.0223 172.1540 0084933 271.3400  99.3297 14.44788026679770
0 COSMOS 2251 DEB
1 34859U 93036ACD 22068.60578052  .00002086  00000-0  51275-3 0  9990
2 34859  74.0397 126.1327 0010510 244.8798 241.1426 14.52914086681614
0 COSMOS 2251 DEB
1 34860U 93036ACE 22069.09428384  .00005755  00000-0  10978-2 0  9998
2 34860  74.0261  49.3940 0020453 118.1211 242.2027 14.64077621684196
0 COSMOS 2251 DEB
1 34861U 93036ACF 22068.77101935  .00003998  00000-0  10566-2 0  9993
2 34861  74.0347 179.6206 0027176 222.7540 260.7676 14.48976749679594
0 IRIDIUM 33 DEB
1 34868U 97051MN  22068.77895273  .00009087  00000-0  17768-2 0  9992
2 34868  86.5382 352.2225 0107495 254.5888 104.3427 14.59111720679396
0 IRIDIUM 33 DEB
1 34869U 97051MP  22068.72814705  .00007079  00000-0  19679-2 0  9991
2 34869  86.4535 335.3342 0049174 108.1732 252.4833 14.45618474674753
0 IRIDIUM 33 DEB
1 34870U 97051MQ  22068.71610280  .00002625  00000-0  65187-3 0  9992
2 34870  86.3251 206.5032 0030884 184.7190 241.7676 14.51221386688547
0 COSMOS 2251 DEB
1 34873U 93036ACL 22068.27020564  .00037094  00000-0  45741-2 0  9996
2 34873  73.9980  21.9949 0018313  36.6800 323.5627 14.82594269690706
0 COSMOS 2251 DEB
1 34874U 93036ACM 22068.71674860  .00001174  00000-0  47832-3 0  9994
2 34874  74.0406  56.3092 0057869  55.0919 117.8141 14.27633105673863
0 COSMOS 2251 DEB
1 34875U 93036ACN 22068.40458675  .00000812  00000-0  25287-3 0  9993
2 34875  74.0479 197.2841 0018144 300.4026 115.4136 14.42450355679023
0 COSMOS 2251 DEB
1 34877U 93036ACQ 22068.78964781  .00013824  00000-0  72243-3 0  9997
2 34877  73.9287  96.9875 0076680 299.7549  59.6031 15.13337057706184
0 COSMOS 2251 DEB
1 34878U 93036ACR 22068.33659903  .00000558  00000-0  58313-3 0  9998
2 34878  73.9653  61.3649 0447702 309.3269 111.3811 13.41006743628292
0 COSMOS 2251 DEB
1 34879U 93036ACS 22066.52949017  .00001121  00000-0  12097-2 0  9991
2 34879  73.9154 277.4643 0535447 240.7236 113.9334 13.25427284400317
0 COSMOS 2251 DEB
1 34882U 93036ACV 22068.96016575  .00000377  00000-0  84836-4 0  9993
2 34882  74.0496 322.8485 0087036  82.9487 278.1565 14.57736761688431
0 COSMOS 2251 DEB
1 34883U 93036ACW 22066.88353996  .00002041  00000-0  92759-3 0  9996
2 34883  74.2105  14.6061 0160878 228.1537 130.5782 14.15757171669616
0 COSMOS 2251 DEB
1 34884U 93036ACX 22069.06015687  .00002215  00000-0  59681-3 0  9992
2 34884  74.0343 168.8255 0018090 233.6168 300.3001 14.48448392677321
0 COSMOS 2251 DEB
1 34885U 93036ACY 22069.19882623  .00001215  00000-0  38312-3 0  9996
2 34885  74.0367 223.8714 0018547 284.8097 249.4069 14.41299105676398
0 COSMOS 2251 DEB
1 34886U 93036ACZ 22068.76433973  .00003051  00000-0  39787-3 0  9996
2 34886  73.9730  83.4252 0105187 245.0350 113.9869 14.77220822424681
0 IRIDIUM 33 DEB
1 34889U 97051MR  22068.70999795  .00012163  00000-0  20237-2 0  9996
2 34889  86.3700 220.0911 0007723 283.4661  76.5694 14.69977168685758
0 IRIDIUM 33 DEB
1 34890U 97051MS  22068.73666893  .00008720  00000-0  20134-2 0  9991
2 34890  86.3038 240.4698 0067983 347.8208  12.1351 14.53715326678216
0 IRIDIUM 33 DEB
1 34896U 97051MY  22068.64473914  .00001949  00000-0  48987-3 0  9990
2 34896  86.3384 212.6715 0036726 180.9026 189.6345 14.50387111676161
0 IRIDIUM 33 DEB
1 34898U 97051NA  22065.25960100  .00008438  00000-0  13895-2 0  9993
2 34898  86.3358 175.6714 0017825 288.7032  71.2249 14.70329153687484
0 IRIDIUM 33 DEB
1 34899U 97051NB  22069.16785208  .00074713  00000-0  58753-2 0  9993
2 34899  86.4028 222.0075 0019463 216.1535 317.8495 15.00358715685705
0 COSMOS 2251 DEB
1 34907U 93036ADC 22068.31847260  .00003884  00000-0  10545-2 0  9993
2 34907  74.0228 193.2817 0020380 227.5354 160.9649 14.47789433677948
0 COSMOS 2251 DEB
1 34908U 93036ADD 22068.31683566  .00000925  00000-0  35856-3 0  9997
2 34908  74.0622  13.4831 0044835  13.2943 148.0878 14.30822193675831
0 COSMOS 2251 DEB
1 34909U 93036ADE 22068.40541348  .00001012  00000-0  17880-3 0  9994
2 34909  74.0167 226.4351 0104973 346.8306  13.0124 14.65343822466521
0 COSMOS 2251 DEB
1 34910U 93036ADF 22068.53520685  .00000292  00000-0  32434-3 0  9992
2 34910  73.8748 103.3655 0477379  95.9505  53.7786 13.34437842633589
0 COSMOS 2251 DEB
1 34911U 93036ADG 22069.18604561  .00000296  00000-0  70057-4 0  9999
2 34911  74.0444 335.2842 0086923  90.7508  59.7073 14.56368134686795
0 COSMOS 2251 DEB
1 34912U 93036ADH 22068.86094139  .00003697  00000-0  54470-3 0  9996
2 34912  74.1635 268.2039 0062190 189.8537 170.1424 14.74413501692596
0 COSMOS 2251 DEB
1 34913U 93036ADJ 22068.91552406  .00000040  00000-0  23738-4 0  9992
2 34913  74.0262 278.7505 0021177 359.6267  63.5326 14.34018747454910
0 COSMOS 2251 DEB
1 34914U 93036ADK 22069.02395420  .00003034  00000-0  46797-3 0  9992
2 34914  74.0020 314.6604 0161548 209.7119 149.4803 14.65419838687834
0 COSMOS 2251 DEB
1 34916U 93036ADM 22069.06378841  .00006082  00000-0  85644-3 0  9996
2 34916  74.0756 294.7348 0094890  80.5134  68.6913 14.74408845694447
0 COSMOS 2251 DEB
1 34918U 93036ADP 22068.52093264  .00002432  00000-0  82560-3 0  9990
2 34918  74.0790 351.4856 0048432 313.9821 220.0259 14.36536698454749
0 COSMOS 2251 DEB
1 34920U 93036ADR 22068.53228534  .00003892  00000-0  98823-3 0  9994
2 34920  74.0327 165.3855 0016366 214.7923 145.2169 14.51030235471943
0 COSMOS 2251 DEB
1 34922U 93036ADT 22068.74246655  .00011961  00000-0  10442-2 0  9999
2 34922  74.1577  66.8940 0090105 312.6067  46.7541 14.93738499393533
0 COSMOS 2251 DEB
1 34924U 93036ADV 22068.51274631  .00001670  00000-0  45445-3 0  9999
2 34924  74.0459 157.5326 0071807 228.1717 143.1222 14.46777584677821
0 IRIDIUM 33 DEB
1 34926U 97051NF  22068.53561420  .00002056  00000-0  94769-3 0  9991
2 34926  86.1524 256.6578 0166777  40.8061  12.0340 14.13521759668451
0 IRIDIUM 33 DEB
1 34928U 97051NH  22067.72368495  .00003660  00000-0  75158-3 0  9993
2 34928  86.4023 232.0839 0033320  76.4952 283.9967 14.60057780459192
0 SES 7 (PROTOSTAR 2)
1 34941U 09027A   22069.10119059 -.00000347  00000-0  00000+0 0  9998
2 34941   0.0269  90.2836 0001596 211.6624  10.4813  1.00271604 46964
0 BREEZE-M R/B
1 34942U 09027B   22068.38374687 -.00000136  00000-0  00000-0 0  9995
2 34942  10.0726   0.6573 3434994 130.3818 343.8846  1.57275646 73606
0 COSMOS 2251 DEB
1 34944U 93036ADX 22068.29006609  .00000856  00000-0  40395-3 0  9998
2 34944  73.7652 176.0547 0236487 248.1356 109.4482 14.07838596448561
0 COSMOS 2251 DEB
1 34946U 93036ADZ 22068.69197341  .00000953  00000-0  24026-3 0  9999
2 34946  74.0231  68.4579 0061240 187.4689 172.5567 14.51426189682145
0 COSMOS 2251 DEB
1 34948U 93036AEB 22069.11219296  .00003869  00000-0  10155-2 0  9999
2 34948  74.0225 191.8834 0026142 221.4006 312.5580 14.49335414679402
0 COSMOS 2251 DEB
1 34949U 93036AEC 22068.75168075  .00000819  00000-0  15395-3 0  9991
2 34949  74.0211 247.1976 0094451  17.8557 342.5890 14.63584686682028
0 COSMOS 2251 DEB
1 34951U 93036AEE 22068.83218087  .00002712  00000-0  47537-3 0  9999
2 34951  74.0172 268.3867 0065085  23.3243 337.0854 14.66912470684161
0 COSMOS 2251 DEB
1 34953U 93036AEG 22068.57031157  .00001219  00000-0  29648-3 0  9997
2 34953  74.0968 115.2875 0076852 120.7890   2.8096 14.52112258680903
0 COSMOS 2251 DEB
1 34954U 93036AEH 22068.57542817  .00003941  00000-0  10120-2 0  9996
2 34954  73.9942 156.7957 0030759 255.3742 104.4007 14.50289758678526
0 COSMOS 2251 DEB
1 34955U 93036AEJ 22068.76560852  .00001622  00000-0  21020-3 0  9993
2 34955  74.0443  97.7840 0118401 174.7214 185.5243 14.76912243696262
0 COSMOS 2251 DEB
1 34958U 93036AEM 22068.33437584  .00003434  00000-0  29258-3 0  9998
2 34958  74.0132 289.5354 0140459  10.1208 350.2747 14.90815822694972
0 COSMOS 2251 DEB
1 34962U 93036AER 22068.17106793  .00000791  00000-0  25909-3 0  9992
2 34962  74.0321 225.0851 0020611 313.0411  46.9017 14.39987225677556
0 COSMOS 2251 DEB
1 34963U 93036AES 22068.27065182  .00001905  00000-0  66492-3 0  9990
2 34963  74.1780  46.8288 0054816 266.9925  92.4954 14.35105892669971
0 COSMOS 2251 DEB
1 34964U 93036AET 22068.69491403  .00001339  00000-0  88065-3 0  9997
2 34964  73.9797 194.5344 0238274 344.2924  15.0896 13.89966196653252
0 COSMOS 2251 DEB
1 34965U 93036AEU 22069.00384100  .00000494  00000-0  15148-3 0  9992
2 34965  74.0547 168.8525 0029131 280.4120  79.3755 14.44020387680524
0 COSMOS 2251 DEB
1 34966U 93036AEV 22069.10131929  .00044781  00000-0  22841-2 0  9997
2 34966  73.9608 331.2475 0027881 190.0952 169.9687 15.16183762697266
0 COSMOS 2251 DEB
1 34967U 93036AEW 22068.84596133  .00000845  00000-0  29864-3 0  9996
2 34967  73.9948 275.8500 0043728  24.1722 336.1468 14.35677831454212
0 COSMOS 2251 DEB
1 34968U 93036AEX 22068.81022019  .00000372  00000-0  86367-4 0  9998
2 34968  74.0420 338.3281 0090307  87.8551 273.2950 14.56173944684606
0 COSMOS 2251 DEB
1 34969U 93036AEY 22069.06426248  .00000419  00000-0  17133-3 0  9995
2 34969  74.0175   4.5842 0077734 100.8127 260.1784 14.28574832674006
0 COSMOS 2251 DEB
1 34971U 93036AFA 22068.53774601  .00000343  00000-0  89251-4 0  9995
2 34971  73.9818   9.4709 0075141 198.5068 279.8038 14.51976470462065
0 COSMOS 2251 DEB
1 34972U 93036AFB 22068.96531394  .00001902  00000-0  51495-3 0  9993
2 34972  73.9961 145.5546 0056164 268.3035 265.8289 14.47505791677431
0 COSMOS 2251 DEB
1 34976U 93036AFF 22068.70056972  .00001857  00000-0  42036-3 0  9993
2 34976  74.0647  59.2967 0065074 110.8031  62.5062 14.55581311681711
0 COSMOS 2251 DEB
1 34977U 93036AFG 22068.35576737  .00005486  00000-0  13342-2 0  9990
2 34977  74.0286 174.8153 0054975 232.0328 127.5860 14.52117011676773
0 COSMOS 2251 DEB
1 34978U 93036AFH 22068.50475811  .00009272  00000-0  11961-2 0  9991
2 34978  74.0249 197.6585 0038412 319.3617  40.4693 14.80594056686927
0 COSMOS 2251 DEB
1 34979U 93036AFJ 22068.93092794  .00001871  00000-0  40992-3 0  9998
2 34979  74.0477  30.5221 0071215 105.0823 255.8235 14.56828609681038
0 IRIDIUM 33 DEB
1 34983U 97051NP  22065.69226414  .00036387  00000-0  48280-2 0  9999
2 34983  86.3569 224.5647 0028855 342.3732  17.6485 14.79361161465137
0 IRIDIUM 33 DEB
1 34985U 97051NR  22068.86570914  .00000742  00000-0  25362-3 0  9993
2 34985  86.3687 278.6826 0025450  56.8557 315.5109 14.34878871674316
0 IRIDIUM 33 DEB
1 34986U 97051NS  22068.47412993  .00008306  00000-0  18606-2 0  9999
2 34986  86.3806 258.2073 0027990 176.2906 183.8515 14.56309757678243
0 IRIDIUM 33 DEB
1 34987U 97051NT  22068.97705843  .00012408  00000-0  13894-2 0  9996
2 34987  86.3378 161.7500 0008579 167.9686 192.1745 14.86757683696241
0 COSMOS 2251 DEB
1 34989U 93036AFM 22068.68891277  .00003106  00000-0  69768-3 0  9999
2 34989  74.0219  58.6373 0030240 180.2165 179.8991 14.56670492678674
0 COSMOS 2251 DEB
1 34990U 93036AFN 22064.30413639  .00000472  00000-0  16692-3 0  9999
2 34990  74.0453 274.2239 0017346 318.7974  41.1869 14.37223666675130
0 COSMOS 2251 DEB
1 34991U 93036AFP 22068.28073979  .00006695  00000-0  13050-2 0  9997
2 34991  74.0468  42.7908 0014345 166.9752 193.1788 14.63131475683021
0 COSMOS 2251 DEB
1 34993U 93036AFR 22068.35045259  .00024331  00000-0  34314-2 0  9990
2 34993  74.0119  53.2793 0021073  73.6907 286.6583 14.76999003683144
0 COSMOS 2251 DEB
1 34995U 93036AFT 22068.34916956  .00007344  00000-0  13361-2 0  9996
2 34995  73.8303 311.9428 0043075 100.8275 259.7745 14.65676007464539
0 COSMOS 2251 DEB
1 34997U 93036AFV 22068.02437255  .00001125  00000-0  41922-3 0  9999
2 34997  73.8895 337.8519 0122650 232.8471 126.1427 14.29151484510179
0 COSMOS 2251 DEB
1 34998U 93036AFW 22069.12534218  .00000838  00000-0  26654-3 0  9994
2 34998  74.0150 200.7680 0023782 313.4493 220.3416 14.41322755676645
0 COSMOS 2251 DEB
1 35000U 93036AFY 22067.94122741  .00003095  00000-0  63958-3 0  9997
2 35000  74.0409  33.3276 0073177 101.9277 259.0105 14.59107926681595
0 COSMOS 2251 DEB
1 35012U 93036AFZ 22069.19296351  .00000243  00000-0  77464-4 0  9998
2 35012  74.0216 149.7999 0084924 263.3372  95.8113 14.43246111677592
0 COSMOS 2251 DEB
1 35014U 93036AGB 22068.93268538  .00001860  00000-0  50351-3 0  9998
2 35014  74.0090 142.4552 0015334 263.4855  96.4559 14.48379539680934
0 COSMOS 2251 DEB
1 35015U 93036AGC 22068.84083213  .00001190  00000-0  40528-3 0  9995
2 35015  73.9509 267.2777 0089173  82.3206 290.7188 14.35355640456631
0 COSMOS 2251 DEB
1 35016U 93036AGD 22068.37521447  .00000499  00000-0  18676-3 0  9990
2 35016  73.9311 299.3811 0121474 151.0522  21.9060 14.30378228671043
0 COSMOS 2251 DEB
1 35018U 93036AGF 22068.64502704  .00000816  00000-0  71122-3 0  9995
2 35018  73.9810 199.5034 0364059  27.9089 334.1032 13.60953861644552
0 COSMOS 2251 DEB
1 35019U 93036AGG 22068.92425543  .00009564  00000-0  12562-2 0  9993
2 35019  74.0073 258.3845 0021367 356.1000   4.0006 14.80194491687484
0 COSMOS 2251 DEB
1 35020U 93036AGH 22068.37568030  .00000456  00000-0  97272-4 0  9997
2 35020  74.0500 302.5813 0096026  50.9651 310.0011 14.59119223681715
0 COSMOS 2251 DEB
1 35022U 93036AGK 22066.69181742  .00000889  00000-0  22082-3 0  9999
2 35022  74.0280  52.1218 0055910 233.9213 125.6763 14.52433984391171
0 COSMOS 2251 DEB
1 35023U 93036AGL 22069.23189994  .00003164  00000-0  65602-3 0  9994
2 35023  74.0553   1.7797 0062589  75.7325  80.9005 14.59398873680877
0 COSMOS 2251 DEB
1 35025U 93036AGN 22069.10494136  .00001014  00000-0  19087-3 0  9997
2 35025  73.9410 209.4011 0097897  78.2615 282.9515 14.62948109685620
0 COSMOS 2251 DEB
1 35026U 93036AGP 22068.62785358  .00000123  00000-0  35761-4 0  9991
2 35026  74.0479  29.6777 0084598 132.4638  40.3802 14.52288502681872
0 COSMOS 2251 DEB
1 35028U 93036AGR 22068.52786552  .00000222  00000-0  56139-4 0  9991
2 35028  74.0412 354.2383 0085005 115.7864  56.8232 14.54653187682923
0 COSMOS 2251 DEB
1 35031U 93036AGU 22068.25424808  .00000233  00000-0  88938-4 0  9992
2 35031  74.0370 266.1872 0019685 337.1676  22.8599 14.36027646455401
0 COSMOS 2251 DEB
1 35032U 93036AGV 22068.85676468  .00001098  00000-0  31414-3 0  9996
2 35032  73.8545  96.2347 0086747 221.1610 312.7685 14.44159596677609
0 COSMOS 2251 DEB
1 35033U 93036AGW 22068.13694198  .00001058  00000-0  33538-3 0  9993
2 35033  74.0308 219.0731 0017105 297.7731  62.1690 14.41226343460690
0 COSMOS 2251 DEB
1 35037U 93036AHA 22069.12968009  .00002653  00000-0  53888-3 0  9994
2 35037  74.0058 344.4551 0043610 126.4164 234.1036 14.61063693683867
0 COSMOS 2251 DEB
1 35038U 93036AHB 22068.43108786  .00001676  00000-0  85563-3 0  9995
2 35038  74.1123 320.6042 0134311 299.3932 236.2003 14.11882574660948
0 COSMOS 2251 DEB
1 35039U 93036AHC 22066.49971160  .00016152  00000-0  23312-2 0  9992
2 35039  74.0155 349.0868 0020572  43.6112 316.6683 14.76098749683426
0 COSMOS 2251 DEB
1 35042U 93036AHF 22068.85373529  .00013136  00000-0  12577-2 0  9993
2 35042  74.0105 111.6046 0030327 263.3473  96.4258 14.92869595691160
0 COSMOS 2251 DEB
1 35043U 93036AHG 22067.83121979  .00003324  00000-0  10769-2 0  9990
2 35043  73.9698 100.4491 0202053 259.6237  98.2082 14.28424854475573
0 COSMOS 2251 DEB
1 35045U 93036AHJ 22068.90908377  .00000602  00000-0  39150-3 0  9998
2 35045  73.7204 309.3045 0190610  36.7528 337.3077 13.96236372654449
0 IRIDIUM 33 DEB
1 35049U 97051NV  22068.88971802  .00008221  00000-0  15642-2 0  9997
2 35049  86.3735 221.7812 0028070 164.4634 195.7446 14.63754129515260
0 IRIDIUM 33 DEB
1 35050U 97051NW  22069.19986387  .00030765  00000-0  28224-2 0  9996
2 35050  86.3008 153.0144 0016252 160.6156 199.5695 14.94634356686828
0 IRIDIUM 33 DEB
1 35051U 97051NX  22068.76750534  .00001059  00000-0  33851-3 0  9999
2 35051  86.3248 242.1818 0007392 311.7713  48.2853 14.38749664671902
0 IRIDIUM 33 DEB
1 35052U 97051NY  22068.75672452  .00000362  00000-0  11811-3 0  9999
2 35052  86.3283 247.7015 0009875 258.5052 101.5036 14.35831672675441
0 IRIDIUM 33 DEB
1 35054U 97051PA  22068.58163215  .00052941  00000-0  49436-2 0  9997
2 35054  86.3332 183.3567 0015039 160.9043 199.2752 14.93843640687039
0 COSMOS 2251 DEB
1 35058U 93036AHM 22068.55965927  .00036723  00000-0  15801-2 0  9996
2 35058  73.9990 258.2649 0036013 317.6460  42.1958 15.21859407697559
0 COSMOS 2251 DEB
1 35059U 93036AHN 22068.19119684  .00008890  00000-0  13868-2 0  9993
2 35059  73.9920 324.1994 0023539  77.8194  79.0963 14.72825558684171
0 COSMOS 2251 DEB
1 35060U 93036AHP 22068.87816981  .00000421  00000-0  15017-3 0  9995
2 35060  74.0368 258.6846 0019558 322.4538  37.5248 14.37198017674791
0 COSMOS 2251 DEB
1 35061U 93036AHQ 22069.20645138  .00001231  00000-0  29074-3 0  9996
2 35061  74.0496  44.3278 0050555 151.1246 209.2732 14.54490131686305
0 COSMOS 2251 DEB
1 35062U 93036AHR 22065.55865222  .00002811  00000-0  14743-2 0  9990
2 35062  74.1010   8.4614 0150036 349.2636  10.5308 14.09046247660750
0 COSMOS 2251 DEB
1 35065U 93036AHU 22068.70876305  .00000733  00000-0  24272-3 0  9997
2 35065  74.0225 224.2495 0021706 320.8996  39.0588 14.39575251485431
0 COSMOS 2251 DEB
1 35066U 93036AHV 22068.00129093  .00005561  00000-0  88257-3 0  9993
2 35066  74.0123 276.8024 0029647  66.9350 293.4943 14.72085109465584
0 COSMOS 2251 DEB
1 35067U 93036AHW 22068.68145541  .00004388  00000-0  79467-3 0  9993
2 35067  73.7560 191.8220 0065711  41.1165 319.4926 14.65239766682922
0 COSMOS 2251 DEB
1 35071U 93036AJA 22067.21268576  .00002725  00000-0  35432-3 0  9997
2 35071  73.9447  48.8699 0113289 272.1769  86.6436 14.76810284470931
0 COSMOS 2251 DEB
1 35072U 93036AJB 22062.85959146  .00006046  00000-0  14248-2 0  9994
2 35072  74.0894 240.8285 0049915 196.2075 163.7494 14.53719034678123
0 COSMOS 2251 DEB
1 35073U 93036AJC 22068.81641210  .00002097  00000-0  51359-3 0  9994
2 35073  74.0201 109.8261 0041265 205.5937 154.3183 14.52615347473186
0 COSMOS 2251 DEB
1 35076U 93036AJF 22068.69596693  .00002771  00000-0  47537-3 0  9999
2 35076  73.9703 222.6952 0069121  43.0851 317.5698 14.67676926421954
0 IRIDIUM 33 DEB
1 35079U 97051PE  22067.14921954  .00006167  00000-0  24520-2 0  9993
2 35079  86.4332  19.3250 0146356 240.8803 117.7670 14.22894076665562
0 IRIDIUM 33 DEB
1 35080U 97051PF  22068.90589860  .00003270  00000-0  13762-2 0  9994
2 35080  86.2376 297.2110 0184858   2.0210 358.1679 14.16753855499502
0 FENGYUN 1C DEB
1 35088U 99025DJQ 22069.05366296  .00001094  00000-0  70222-3 0  9995
2 35088  94.6561 191.7189 0355518   7.9094 352.7472 13.77010945711083
0 FENGYUN 1C DEB
1 35089U 99025DJR 22068.83888597  .00000636  00000-0  35052-3 0  9997
2 35089  98.6590 107.9133 0030569  62.4532 297.9742 14.14778756414215
0 FENGYUN 1C DEB
1 35090U 99025DJS 22066.33108209  .00001585  00000-0  97196-3 0  9998
2 35090  99.1314  33.4709 0062523 270.0541  89.3462 14.06307285468386
0 FENGYUN 1C DEB
1 35091U 99025DJT 22068.41487707  .00002402  00000-0  88846-3 0  9994
2 35091  98.6486 142.9762 0079261  93.4810 267.5443 14.31534103701152
0 FENGYUN 1C DEB
1 35093U 99025DJV 22069.09313792  .00007089  00000-0  14779-2 0  9992
2 35093  98.6130 179.2810 0152171 143.3970 217.7739 14.53197401716714
0 FENGYUN 1C DEB
1 35094U 99025DJW 22067.27696788  .00016945  00000-0  33845-2 0  9996
2 35094  97.7170 255.6987 0174996 197.7420 336.5790 14.52648999456838
0 FENGYUN 1C DEB
1 35095U 99025DJX 22068.10746513  .00004679  00000-0  87494-3 0  9993
2 35095  98.4651 171.5719 0210410  75.1226 287.3168 14.52068839517812
0 FENGYUN 1C DEB
1 35097U 99025DJZ 22069.21361893  .00004108  00000-0  43620-3 0  9992
2 35097  98.6896  56.5069 0174592 192.5676 167.1153 14.79098715743679
0 FENGYUN 1C DEB
1 35099U 99025DKB 22067.39600205  .00025060  00000-0  13347-2 0  9996
2 35099  98.3438 292.7078 0122134 359.1384   0.9613 15.08908717751361
0 FENGYUN 1C DEB
1 35102U 99025DKE 22068.34629292  .00000619  00000-0  28270-3 0  9999
2 35102  98.5769 100.3067 0065141 152.1019 208.3678 14.23867520453251
0 FENGYUN 1C DEB
1 35104U 99025DKG 22068.37017294  .00000441  00000-0  24087-3 0  9992
2 35104  98.6732  97.5722 0033796 138.3579 233.9498 14.16676377702865
0 FENGYUN 1C DEB
1 35105U 99025DKH 22069.01429697  .00000353  00000-0  11787-3 0  9994
2 35105  98.8365 340.0271 0121717 178.3889 249.6772 14.39805773746916
0 FENGYUN 1C DEB
1 35108U 99025DKL 22067.81144466  .00000193  00000-0  13641-3 0  9994
2 35108  98.8065  88.2377 0137075 321.7604  37.3902 14.04584092699236
0 FENGYUN 1C DEB
1 35109U 99025DKM 22068.28129935  .00000380  00000-0  18095-3 0  9997
2 35109  98.5788  83.3289 0159011  46.4171 315.0078 14.18223097446621
0 FENGYUN 1C DEB
1 35110U 99025DKN 22064.37487103  .00003075  00000-0  79902-3 0  9997
2 35110  96.6636 306.5795 0121756 154.4537 206.2745 14.45793473741357
0 FENGYUN 1C DEB
1 35111U 99025DKP 22069.22722590  .00000554  00000-0  35355-3 0  9992
2 35111  99.5471  53.0464 0024437  55.4390 304.9078 14.08058474733696
0 FENGYUN 1C DEB
1 35114U 99025DKS 22068.28062862  .00026694  00000-0  96409-3 0  9999
2 35114  97.7950 268.8881 0129186 142.2025 218.8371 15.20828287476186
0 FENGYUN 1C DEB
1 35116U 99025DKU 22068.87704594  .00001015  00000-0  51476-3 0  9996
2 35116  98.7852 117.7345 0067697  51.3395 309.3811 14.17037155404542
0 FENGYUN 1C DEB
1 35117U 99025DKV 22069.08802145  .00003604  00000-0  11806-2 0  9990
2 35117  98.7481 176.1122 0050492  34.9652  91.1818 14.38325842695946
0 FENGYUN 1C DEB
1 35118U 99025DKW 22065.65179339  .00004961  00000-0  66016-3 0  9993
2 35118  97.9537 208.8552 0169844 327.6695  31.4190 14.70574253760931
0 FENGYUN 1C DEB
1 35119U 99025DKX 22058.11084668  .00006131  00000-0  10168-2 0  9994
2 35119  99.9613  21.1208 0044792 156.4840 261.4251 14.70038957453249
0 FENGYUN 1C DEB
1 35120U 99025DKY 22064.94859220  .00003051  00000-0  17282-2 0  9990
2 35120  98.0161 123.7215 0159164 343.3745  16.2240 14.04505450443514
0 FENGYUN 1C DEB
1 35121U 99025DKZ 22061.10401992  .00000302  00000-0  19415-3 0  9992
2 35121  98.3330  18.5127 0161266  53.8432 307.7535 14.03228735417053
0 FENGYUN 1C DEB
1 35122U 99025DLA 22068.37047733  .00000307  00000-0  19092-3 0  9994
2 35122  98.8062 104.4335 0026984 109.3719 251.0373 14.11883350444600
0 FENGYUN 1C DEB
1 35123U 99025DLB 22068.61724390  .00000686  00000-0  34984-3 0  9990
2 35123  98.9923 199.2449 0055234  40.2680 320.2566 14.18336312415593
0 FENGYUN 1C DEB
1 35125U 99025DLD 22067.90253448  .00001238  00000-0  56491-3 0  9995
2 35125  98.3233 138.4669 0061734  31.8052 328.6832 14.22230873730311
0 FENGYUN 1C DEB
1 35128U 99025DLG 22069.04713344  .00003506  00000-0  18954-2 0  9992
2 35128  98.3604   4.6113 0134366  27.7739 333.0524 14.08652579712281
0 FENGYUN 1C DEB
1 35130U 99025DLJ 22068.00229873  .00003814  00000-0  18159-2 0  9998
2 35130  99.0625 172.8674 0053678 193.0301 166.9490 14.19198425451061
0 FENGYUN 1C DEB
1 35134U 99025DLN 22063.46238013  .00003551  00000-0  92524-3 0  9990
2 35134  98.8729 319.2805 0063654 302.8600  56.6475 14.49023676695748
0 FENGYUN 1C DEB
1 35135U 99025DLP 22067.03129295  .00000502  00000-0  25542-3 0  9998
2 35135  98.6540  97.8763 0085496 302.0447  87.2012 14.18349153682985
0 FENGYUN 1C DEB
1 35136U 99025DLQ 22067.46050197  .00002803  00000-0  99010-3 0  9997
2 35136  98.6849 151.9990 0075872  68.2191 292.7045 14.33854380386302
0 FENGYUN 1C DEB
1 35137U 99025DLR 22068.67790303  .00009949  00000-0  31567-2 0  9992
2 35137  99.0915 220.4697 0233857 210.6093 162.8066 14.25735991492508
0 FENGYUN 1C DEB
1 35139U 99025DLT 22068.81194415  .00000370  00000-0  19778-3 0  9992
2 35139  98.6533  96.8446 0070482 302.7954  56.6445 14.17836947446650
0 FENGYUN 1C DEB
1 35140U 99025DLU 22067.82704656  .00000238  00000-0  88223-4 0  9993
2 35140  98.4365  98.0710 0299241   8.0031 165.9966 14.18968352677665
0 FENGYUN 1C DEB
1 35141U 99025DLV 22068.97031934  .00001374  00000-0  80558-3 0  9997
2 35141  99.0579 148.3120 0076958 314.6533 221.4263 14.08467122710568
0 FENGYUN 1C DEB
1 35142U 99025DLW 22068.78912779  .00000149  00000-0  15034-3 0  9998
2 35142  99.0292  86.1078 0109226 345.4076 190.5469 13.89368235666030
0 FENGYUN 1C DEB
1 35144U 99025DLY 22066.15693662  .00002789  00000-0  10785-2 0  9995
2 35144  98.7304 141.1228 0046994 145.8461 214.5760 14.30308845453913
0 FENGYUN 1C DEB
1 35145U 99025DLZ 22066.61397838  .00019640  00000-0  90427-3 0  9999
2 35145  99.6013 204.0695 0087882  20.4419 340.0289 15.16931985767696
0 FENGYUN 1C DEB
1 35146U 99025DMA 22068.03214953  .00007976  00000-0  13328-2 0  9997
2 35146  98.2498 179.2687 0245521 114.0108 248.6997 14.52121518447162
0 FENGYUN 1C DEB
1 35147U 99025DMB 22068.40328160  .00002367  00000-0  76152-3 0  9999
2 35147  99.5093 305.8561 0049176 303.7547  55.8957 14.39742012543577
0 FENGYUN 1C DEB
1 35148U 99025DMC 22066.31145043  .00004710  00000-0  98047-3 0  9996
2 35148  98.9734  88.2614 0088014   7.1588 353.0843 14.58141945763781
0 FENGYUN 1C DEB
1 35152U 99025DMG 22068.59999095  .00000768  00000-0  76216-3 0  9990
2 35152  99.7056 201.8786 0198496 327.7687  31.1454 13.73357494716562
0 FENGYUN 1C DEB
1 35153U 99025DMH 22068.55157504  .00000948  00000-0  15234-3 0  9991
2 35153  98.8869 180.7704 0202906   9.4286 351.0632 14.61696449770703
0 FENGYUN 1C DEB
1 35154U 99025DMJ 22066.78018983  .00000409  00000-0  20105-3 0  9997
2 35154  99.5003  89.0904 0186826 299.8101  58.4609 14.15128927450952
0 FENGYUN 1C DEB
1 35155U 99025DMK 22068.85970894  .00000555  00000-0  30937-3 0  9992
2 35155  98.6260 117.3701 0028074  67.8312 292.5837 14.14612699736724
0 FENGYUN 1C DEB
1 35156U 99025DML 22068.75189276  .00000451  00000-0  28698-3 0  9994
2 35156  98.7057  83.8400 0038785  79.0907 281.4623 14.08161187446922
0 FENGYUN 1C DEB
1 35159U 99025DMP 22068.04128769  .00001842  00000-0  80412-3 0  9996
2 35159  98.9260  93.1772 0012879 231.6048 156.2520 14.25121830417931
0 FENGYUN 1C DEB
1 35160U 99025DMQ 22068.75822108  .00000478  00000-0  16712-3 0  9990
2 35160  99.2501  83.6740 0186894 269.0768  88.8996 14.31281227452500
0 FENGYUN 1C DEB
1 35161U 99025DMR 22068.28670489  .00008279  00000-0  10350-2 0  9990
2 35161  99.5624 270.4377 0100946 260.0324  98.9477 14.79085895521187
0 FENGYUN 1C DEB
1 35163U 99025DMT 22064.11111380  .00001701  00000-0  37326-3 0  9996
2 35163  99.0989 200.3152 0117148  39.9894 320.9855 14.54960230484902
0 FENGYUN 1C DEB
1 35164U 99025DMU 22063.53879915  .00001552  00000-0  12548-2 0  9999
2 35164  98.6282 326.9183 0213772  18.2725 342.5963 13.81538652737949
0 FENGYUN 1C DEB
1 35165U 99025DMV 22066.08868601  .00001300  00000-0  66408-3 0  9999
2 35165  98.9499 124.4576 0029211  73.7484 286.6901 14.17193896701321
0 FENGYUN 1C DEB
1 35167U 99025DMX 22065.33408160  .00000853  00000-0  28734-3 0  9990
2 35167  98.4588  90.2773 0251148 292.4526  65.0232 14.23444947503801
0 FENGYUN 1C DEB
1 35168U 99025DMY 22057.16361651  .00000211  00000-0  12391-3 0  9996
2 35168  98.2078  35.4591 0130287 170.2477 190.1272 14.12748510383431
0 FENGYUN 1C DEB
1 35169U 99025DMZ 22068.28190381  .00001336  00000-0  31935-3 0  9996
2 35169  99.1033 173.6128 0182589 263.0613  94.9777 14.45984193458405
0 FENGYUN 1C DEB
1 35170U 99025DNA 22065.55733366  .00002388  00000-0  14371-2 0  9993
2 35170  99.8381 188.4566 0107001 115.2556 245.9743 14.05050839419179
0 FENGYUN 1C DEB
1 35171U 99025DNB 22068.45517571  .00001254  00000-0  10336-2 0  9992
2 35171  98.9266 253.9636 0154329  95.9542 265.9227 13.85536299694723
0 FENGYUN 1C DEB
1 35173U 99025DND 22068.02952559  .00002580  00000-0  27215-2 0  9993
2 35173  99.8040 177.8517 0515157 289.8084  64.8213 13.29475303475131
0 FENGYUN 1C DEB
1 35174U 99025DNE 22068.28949731  .00003413  00000-0  58991-3 0  9992
2 35174  99.0512 275.6440 0355791  48.4201 314.6908 14.35145473701405
0 FENGYUN 1C DEB
1 35175U 99025DNF 22068.36965535  .00000326  00000-0  17604-3 0  9998
2 35175  98.6609  98.6222 0059071 289.0701  70.4084 14.18436966417350
0 FENGYUN 1C DEB
1 35176U 99025DNG 22068.83074300  .00000294  00000-0  17316-3 0  9996
2 35176  98.5607  97.8904 0056627  45.6484 314.9309 14.14190357720073
0 FENGYUN 1C DEB
1 35177U 99025DNH 22055.44097315  .00000932  00000-0  48203-3 0  9999
2 35177  98.7324 130.2120 0043229 111.5455 311.5574 14.16880888445572
0 FENGYUN 1C DEB
1 35180U 99025DNL 22068.45929604  .00002688  00000-0  14894-2 0  9991
2 35180  99.5782 230.4633 0617052 256.3687  96.8037 13.43812842362064
0 FENGYUN 1C DEB
1 35183U 99025DNP 22068.33399858  .00003455  00000-0  61731-3 0  9999
2 35183  98.9038  96.7645 0099822 169.6222 190.7071 14.64456111460899
0 FENGYUN 1C DEB
1 35185U 99025DNR 22068.87062030  .00007212  00000-0  81640-3 0  9995
2 35185  99.5579 287.7633 0109909 190.0427 169.8585 14.82506228753730
0 FENGYUN 1C DEB
1 35186U 99025DNS 22060.46321959  .00000906  00000-0  35828-3 0  9991
2 35186  99.6040 293.6792 0063325 129.4691 231.2114 14.30701925450827
0 FENGYUN 1C DEB
1 35189U 99025DNV 22068.34876922  .00001359  00000-0  86074-3 0  9996
2 35189  98.9602 118.9014 0064547  22.7829 337.6180 14.04674335442748
0 FENGYUN 1C DEB
1 35190U 99025DNW 22068.36079114  .00003921  00000-0  10611-2 0  9998
2 35190  98.7013 123.8134 0109841 164.5967 195.8611 14.44788257387000
0 FENGYUN 1C DEB
1 35191U 99025DNX 22068.81634829  .00000150  00000-0  14910-3 0  9997
2 35191  98.9503  82.6375 0103855 351.1417   8.7907 13.90130687688067
0 FENGYUN 1C DEB
1 35192U 99025DNY 22067.73466034  .00007099  00000-0  14056-2 0  9998
2 35192  99.1774  67.3265 0216048 270.1119  87.5317 14.48880698456068
0 FENGYUN 1C DEB
1 35194U 99025DPA 22064.17308798  .00001347  00000-0  35536-3 0  9994
2 35194  98.9948  34.8591 0144417 259.6794  98.8088 14.44847151711991
0 FENGYUN 1C DEB
1 35195U 99025DPB 22068.76325568  .00001106  00000-0  31494-3 0  9990
2 35195  99.0557  51.8924 0126330 214.0243 145.2798 14.42992374710379
0 FENGYUN 1C DEB
1 35196U 99025DPC 22068.90370769  .00000679  00000-0  30345-3 0  9998
2 35196  98.6646 115.5666 0232181 325.2361 209.7567 14.12640030696967
0 FENGYUN 1C DEB
1 35199U 99025DPF 22065.94457101  .00001073  00000-0  83140-3 0  9998
2 35199  98.4522 320.8259 0127382  78.0230 283.5177 13.90799902730070
0 FENGYUN 1C DEB
1 35200U 99025DPG 22068.56648472  .00000237  00000-0  34680-3 0  9991
2 35200  98.9102 192.3971 0354439 275.6515 141.5223 13.40808844474814
0 FENGYUN 1C DEB
1 35201U 99025DPH 22068.87983098  .00000049  00000-0  52166-4 0  9993
2 35201  98.8017  98.7540 0006959 175.2945 358.7752 14.10272395676506
0 FENGYUN 1C DEB
1 35202U 99025DPJ 22068.81771215  .00000507  00000-0  76071-3 0  9993
2 35202 100.2078 109.1413 0381450  39.1705 323.6379 13.32139320386677
0 FENGYUN 1C DEB
1 35203U 99025DPK 22068.10260943  .00001179  00000-0  14915-2 0  9991
2 35203 102.0490 166.2389 0547421 250.6942 103.4205 13.16834871414991
0 FENGYUN 1C DEB
1 35204U 99025DPL 22068.28223984  .00002074  00000-0  93255-3 0  9995
2 35204  98.9330 191.1564 0056404  19.0214  14.0098 14.22618557712497
0 FENGYUN 1C DEB
1 35205U 99025DPM 22067.47219897  .00003371  00000-0  16215-2 0  9995
2 35205  99.0297  72.9366 0068674  95.9807 264.9199 14.18248554407985
0 FENGYUN 1C DEB
1 35207U 99025DPP 22068.11099087  .00001756  00000-0  63945-3 0  9994
2 35207  98.7835 121.5789 0078110  35.7632 324.8747 14.32746495451050
0 FENGYUN 1C DEB
1 35208U 99025DPQ 22067.90265398  .00000839  00000-0  63946-3 0  9993
2 35208  99.0181 107.3478 0105526 338.5783  21.0978 13.93821880713768
0 FENGYUN 1C DEB
1 35210U 99025DPS 22068.34614017  .00002061  00000-0  11528-2 0  9998
2 35210  99.0343  22.5203 0068357 195.3854 317.5260 14.10726858706187
0 FENGYUN 1C DEB
1 35211U 99025DPT 22069.09969060  .00001182  00000-0  79707-3 0  9990
2 35211  98.4511   5.0693 0141638 176.2363 197.8800 13.97496866703570
0 FENGYUN 1C DEB
1 35213U 99025DPV 22068.69235131  .00009742  00000-0  88524-3 0  9999
2 35213  98.4445 332.2727 0153994  90.3172 271.5689 14.86789777721629
0 FENGYUN 1C DEB
1 35214U 99025DPW 22068.31668345  .00017436  00000-0  20807-2 0  9994
2 35214  98.5838 270.9835 0079952 258.7882 275.6863 14.82030101754434
0 FENGYUN 1C DEB
1 35217U 99025DPZ 22068.21412593  .00000951  00000-0  11696-2 0  9998
2 35217 101.8857  55.4043 0319406 211.6616 146.4943 13.49458299426329
0 FENGYUN 1C DEB
1 35219U 99025DQB 22069.15163646  .00014914  00000-0  22118-2 0  9991
2 35219  99.1120 207.2272 0086807 124.0473 236.9012 14.72626743779262
0 FENGYUN 1C DEB
1 35220U 99025DQC 22067.84521803  .00001430  00000-0  80810-3 0  9993
2 35220  99.1367  95.5214 0051478 237.5466 122.0721 14.11212483437267
0 FENGYUN 1C DEB
1 35221U 99025DQD 22066.92198647  .00003177  00000-0  12272-2 0  9994
2 35221  98.4076 229.6776 0072628  12.7788 347.5214 14.29379384716376
0 FENGYUN 1C DEB
1 35222U 99025DQE 22067.69122271  .00008059  00000-0  53450-3 0  9998
2 35222  98.9874  50.8020 0124823 199.5489 160.0922 15.01368703763179
0 FENGYUN 1C DEB
1 35223U 99025DQF 22068.75792689  .00001515  00000-0  21788-3 0  9992
2 35223  98.9968 246.4696 0179672 307.7466  50.7572 14.67888015736154
0 FENGYUN 1C DEB
1 35225U 99025DQH 22068.69819366  .00003585  00000-0  16269-2 0  9995
2 35225 103.2888 221.6235 0052815  31.6138 328.8177 14.22280779409554
0 FENGYUN 1C DEB
1 35226U 99025DQJ 22068.43435099  .00037426  00000-0  30124-2 0  9992
2 35226  98.9828 243.1107 0036410  26.9985 333.3118 14.99326502806091
0 FENGYUN 1C DEB
1 35227U 99025DQK 22068.35064949  .00005304  00000-0  91696-3 0  9998
2 35227  98.9226 117.1700 0074006 235.6781 123.7402 14.67075977463402
0 FENGYUN 1C DEB
1 35229U 99025DQM 22068.01878851  .00002548  00000-0  12299-2 0  9996
2 35229  98.5301 158.1230 0087894 318.8992  40.5585 14.17468849448280
0 FENGYUN 1C DEB
1 35230U 99025DQN 22066.37929304  .00003609  00000-0  13690-2 0  9990
2 35230  98.4000 222.0047 0064617 341.3437  18.5378 14.30529628695396
0 FENGYUN 1C DEB
1 35231U 99025DQP 22068.06364490 -.00000041  00000-0  19032-5 0  9995
2 35231  98.9202 112.1666 0037642  37.5150 322.8639 14.14360265720691
0 FENGYUN 1C DEB
1 35233U 99025DQR 22066.13357711  .00001867  00000-0  84072-3 0  9991
2 35233  98.8252 135.2821 0052812 328.4527  61.1090 14.22767782714996
0 FENGYUN 1C DEB
1 35234U 99025DQS 22069.07950178  .00001663  00000-0  41078-3 0  9997
2 35234  98.2112 115.7620 0134935 203.2124 156.2940 14.48022145743702
0 FENGYUN 1C DEB
1 35236U 99025DQU 22068.84122632  .00000701  00000-0  23825-3 0  9991
2 35236  98.9033 295.5544 0096234 310.4373  48.8448 14.37437466783270
0 FENGYUN 1C DEB
1 35237U 99025DQV 22069.17158141  .00009246  00000-0  14994-2 0  9992
2 35237  99.5946 147.0483 0055559  41.0775 319.4589 14.70446744775205
0 FENGYUN 1C DEB
1 35239U 99025DQX 22068.00995920  .00000420  00000-0  47360-3 0  9998
2 35239  99.2438 274.1108 0236017 340.8784  18.3612 13.64934113393501
0 FENGYUN 1C DEB
1 35240U 99025DQY 22068.09629267  .00001771  00000-0  10188-2 0  9994
2 35240  99.3543 311.2239 0058314  45.2485 315.3408 14.09782512737212
0 FENGYUN 1C DEB
1 35241U 99025DQZ 22068.78532247  .00001387  00000-0  10140-2 0  9993
2 35241  98.5651  70.5093 0153825  66.3099 295.4119 13.91957253684343
0 FENGYUN 1C DEB
1 35242U 99025DRA 22068.55208188  .00000525  00000-0  69906-3 0  9991
2 35242 100.0447  99.3015 0310815 327.3463 188.6476 13.47345756682255
0 FENGYUN 1C DEB
1 35243U 99025DRB 22069.09485794  .00001939  00000-0  46331-3 0  9999
2 35243  98.8971  19.7313 0122997 161.9984 198.5623 14.50399437709031
0 FENGYUN 1C DEB
1 35244U 99025DRC 22067.38293394  .00000012  00000-0  35069-4 0  9991
2 35244  98.9954 115.7707 0074556  38.0776 322.5626 14.02998696465140
0 FENGYUN 1C DEB
1 35245U 99025DRD 22066.95808076  .00001686  00000-0  11008-2 0  9996
2 35245  99.1649 124.9843 0069191 143.4616 341.8452 14.02546855684310
0 FENGYUN 1C DEB
1 35247U 99025DRF 22067.37151051  .00001113  00000-0  46209-3 0  9996
2 35247  98.7794 122.7580 0066270 137.3552 223.2800 14.27291220806066
0 CZ-4C DEB
1 35248U 07055C   22068.48332589  .00002349  00000-0  38171-3 0  9990
2 35248  97.7771 251.8525 0150231  30.5802  22.1434 14.64746120756672
0 CZ-4C DEB
1 35249U 07055D   22069.04165421  .00002014  00000-0  51632-3 0  9999
2 35249  97.8697 352.1269 0242749 314.7673  43.3988 14.35201401745426
0 CZ-4C DEB
1 35251U 07055F   22068.99215883  .00002001  00000-0  45230-3 0  9996
2 35251  97.7822  76.4263 0202499  73.7657 316.1278 14.45294001750732
0 METEOR 2-5 DEB
1 35253U 79095U   22068.33454918  .00006783  00000-0  17394-2 0  9993
2 35253  81.1883  24.0016 0012664 123.3048  30.1797 14.50182609940749
0 METEOR 2-5 DEB
1 35254U 79095V   22066.58534493  .00002924  00000-0  85091-3 0  9999
2 35254  81.2019  28.0814 0010154 154.6091 205.5599 14.44041018446941
0 METEOR 2-5 DEB
1 35255U 79095W   22068.78691735  .00002401  00000-0  10923-2 0  9998
2 35255  81.2063 245.0885 0006871 118.6194 241.5675 14.21470971451158
0 METEOR 2-5 DEB
1 35256U 79095X   22068.20997682  .00003123  00000-0  14199-2 0  9997
2 35256  81.1915 242.7585 0014491 100.6792 259.6017 14.21527323425040
0 METEOR 2-5 DEB
1 35257U 79095Y   22069.03009247  .00003923  00000-0  13987-2 0  9993
2 35257  81.2053 180.0776 0016062  24.0884 336.1048 14.34007316702017
0 ERS 1 DEB
1 35258U 91050M   22068.86455637  .00008717  00000-0  11174-2 0  9999
2 35258  98.2444 122.8321 0016175  12.5761 347.5851 14.81431838714755
0 ERS 1 DEB
1 35259U 91050N   22068.89924964  .00001393  00000-0  37855-3 0  9995
2 35259  98.5351  58.6775 0018967 330.9726  85.5819 14.49025793947961
0 ERS 1 DEB
1 35260U 91050P   22068.93216497  .00001307  00000-0  37464-3 0  9998
2 35260  98.5859  59.0941 0027540  41.9663  10.6301 14.46445206868860
0 ERS 1 DEB
1 35261U 91050Q   22068.85339170  .00008386  00000-0  10524-2 0  9995
2 35261  98.2419 124.5971 0016619 346.7180  13.3592 14.82330986717500
0 ERS 1 DEB
1 35262U 91050R   22068.91844389  .00001550  00000-0  39083-3 0  9994
2 35262  98.3834  44.2333 0019857 194.5031 231.1497 14.52403847611325
0 CZ-4 DEB
1 35268U 04035L   22067.51397840  .00013615  00000-0  54807-3 0  9993
2 35268  97.5635 161.4827 0003444 184.3487 175.7721 15.24995711445546
0 CZ-4 DEB
1 35269U 04035M   22068.56814058  .00003225  00000-0  47542-3 0  9999
2 35269  98.0309 189.6985 0140883 292.5155  66.1189 14.69504651924596
0 CZ-4 DEB
1 35270U 04035N   22068.86638794  .00001894  00000-0  33470-3 0  9990
2 35270  97.7700  32.1679 0174917 173.9697 242.4221 14.58994154920691
0 CZ-2D DEB
1 35271U 05024K   22068.86585587  .00002412  00000-0  12274-3 0  9997
2 35271  97.3396 107.1398 0011717 355.2573   4.8548 15.17704793865623
0 CZ-2D DEB
1 35272U 05024L   22068.85883946  .00002491  00000-0  12460-3 0  9991
2 35272  97.3298 105.8657 0011964 340.3408 193.2230 15.18280669899703
0 CZ-2D DEB
1 35273U 05024M   22068.35914474  .00003228  00000-0  15439-3 0  9996
2 35273  97.3214 109.0549 0011875 314.7366  45.2902 15.19631327938299
0 CZ-2D DEB
1 35274U 08056D   22068.80843090  .00000566  00000-0  32490-3 0  9998
2 35274  99.1285  79.1387 0168114 286.6778  71.5996 14.06722674677979
0 CZ-2D DEB
1 35275U 08056E   22068.48794021  .00000595  00000-0  24349-3 0  9998
2 35275  98.8424  84.5826 0139050 155.2017 205.5930 14.25810852685662
0 COSMOS 2251 DEB
1 35278U 93036AJG 22068.45727313  .00007400  00000-0  17992-2 0  9997
2 35278  74.0474 214.9361 0061121 241.2081 174.5347 14.51831981676379
0 COSMOS 2251 DEB
1 35279U 93036AJH 22068.64010552  .00010309  00000-0  12982-2 0  9995
2 35279  74.0023 193.9219 0031019 315.7955  44.0744 14.81765852685501
0 COSMOS 2251 DEB
1 35280U 93036AJJ 22069.15172360  .00003251  00000-0  87974-3 0  9992
2 35280  74.0329 213.7719 0062577 274.3017 259.8500 14.46990607675616
0 COSMOS 2251 DEB
1 35281U 93036AJK 22068.70382249  .00004401  00000-0  96962-3 0  9997
2 35281  74.0029  63.8699 0019987 319.9850  39.9841 14.57593063679563
0 COSMOS 2251 DEB
1 35282U 93036AJL 22068.99244913  .00003717  00000-0  82612-3 0  9990
2 35282  74.0992 173.8895 0092605 163.0174 197.4126 14.54718716677934
0 COSMOS 2251 DEB
1 35283U 93036AJM 22069.40486883  .02304340  00000-0  14491-1 0  9998
2 35283  73.9691 123.3441 0017604 323.1745  36.8269 15.73158626691632
0 COSMOS 2251 DEB
1 35284U 93036AJN 22068.83609821  .00000212  00000-0  81626-4 0  9990
2 35284  74.0475 268.8997 0017416 326.9234  44.9632 14.36118384457900
0 COSMOS 2251 DEB
1 35285U 93036AJP 22066.79423301  .00003364  00000-0  80525-3 0  9998
2 35285  74.0264 118.4214 0016703 207.5955 152.4323 14.53867815390863
0 COSMOS 2251 DEB
1 35287U 93036AJR 22068.82398205  .00000160  00000-0  65881-4 0  9990
2 35287  74.0288 276.7769 0018512 346.9220  25.0133 14.34679311671796
0 COSMOS 2251 DEB
1 35288U 93036AJS 22068.83260788  .00001387  00000-0  37812-3 0  9990
2 35288  73.9539 107.5519 0050482 290.7801  68.7953 14.47634378679858
0 COSMOS 2251 DEB
1 35290U 93036AJU 22066.69695121  .00002715  00000-0  73890-3 0  9990
2 35290  74.0197 166.3337 0013848 246.3226 113.6480 14.47892779678841
0 COSMOS 2251 DEB
1 35291U 93036AJV 22069.03419187  .00004733  00000-0  36883-3 0  9991
2 35291  74.0401 276.4503 0087234 338.7401 174.0998 14.98497901698873
0 COSMOS 2251 DEB
1 35292U 93036AJW 22068.54565377  .00048856  00000-0  27246-2 0  9996
2 35292  73.9896  10.4555 0020699 133.3206 226.9718 15.13151136693571
0 IRIDIUM 33 DEB
1 35294U 97051PJ  22068.68915691  .00007388  00000-0  15240-2 0  9998
2 35294  86.3314 207.3442 0012704  29.2445  33.7247 14.60287780677405
0 IRIDIUM 33 DEB
1 35296U 97051PL  22068.63547583  .00005400  00000-0  12922-2 0  9993
2 35296  86.3159 209.0354 0005556  92.0145 268.1698 14.53422975510675
0 IRIDIUM 33 DEB
1 35297U 97051PM  22068.34959765  .00000864  00000-0  28624-3 0  9997
2 35297  86.3855 284.0463 0016973 355.2877 129.9089 14.36697237506448
0 IRIDIUM 33 DEB
1 35299U 97051PP  22068.55845528  .00001119  00000-0  43293-3 0  9996
2 35299  86.2903 263.5231 0059593 253.1029 161.9247 14.28325817666440
0 CZ-4 DEB
1 35319U 99057RX  22068.28109466  .00000410  00000-0  21661-3 0  9999
2 35319  98.6758 179.3111 0203004 101.0258 315.8441 14.08604572 44871
0 CZ-4 DEB
1 35320U 99057RY  22068.14757573  .00001480  00000-0  27561-3 0  9997
2 35320  98.2031  42.6265 0002113  75.4504 284.6933 14.66548474815812
0 CZ-4 DEB
1 35322U 99057SA  22068.25728283  .00000279  00000-0  14689-3 0  9999
2 35322  99.1639  59.3888 0202939 162.9038 197.9103 14.11807675450716
0 CZ-4 DEB
1 35323U 99057SB  22068.82238621  .00004936  00000-0  10187-2 0  9997
2 35323  98.3323 100.6846 0073742 104.9626 255.9752 14.59148794461123
0 CZ-4 DEB
1 35325U 99057SD  22068.41894514  .00002999  00000-0  42503-3 0  9998
2 35325  98.4711 121.3947 0016556 314.2029  45.7819 14.77758495470069
0 CZ-4 DEB
1 35326U 99057SE  22069.12832244  .00000143  00000-0  10010-3 0  9997
2 35326  98.5015  18.0696 0227944  11.2337 349.3797 13.99599213881655
0 CZ-4 DEB
1 35327U 99057SF  22067.89191615  .00001094  00000-0  28601-3 0  9991
2 35327  98.6250 103.6079 0091069 220.0384 139.4072 14.49114006952992
0 CZ-4 DEB
1 35329U 99057SH  22068.74406431  .00000067  00000-0  40203-4 0  9995
2 35329  98.2113 249.6403 0134530  31.8128  31.3800 14.27276206 23035
0 CZ-4 DEB
1 35330U 99057SJ  22068.27194040  .00000600  00000-0  19708-3 0  9993
2 35330  98.7485  64.1671 0085350 357.0987   2.9694 14.40030388995769
0 CZ-4 DEB
1 35331U 99057SK  22069.07615734  .00002055  00000-0  52158-3 0  9993
2 35331  98.6552 111.0651 0115211  85.7170 304.8426 14.47981204  8393
0 CZ-4 DEB
1 35332U 99057SL  22066.86033228  .00000063  00000-0  39495-4 0  9997
2 35332  98.3322 298.1354 0124016 103.1700 258.3352 14.26424933996935
0 CZ-4 DEB
1 35334U 99057SN  22068.79353870  .00001099  00000-0  28864-3 0  9998
2 35334  98.5982  86.2735 0076048 113.3219 247.6000 14.49539493841774
0 PSLV DEB
1 35336U 01049PU  22068.87218802  .00008653  00000-0  13683-2 0  9993
2 35336  97.8075 121.6342 0070829 119.9752 240.8510 14.70832341924634
0 PSLV DEB
1 35339U 01049PX  22068.65403185  .00002864  00000-0  80336-3 0  9990
2 35339  98.2291 209.5690 0185468 197.5658 161.9060 14.37126314 13087
0 PSLV DEB
1 35341U 01049PZ  22069.19839136  .00003970  00000-0  46669-3 0  9992
2 35341  97.8278 337.6188 0031734 305.5738  54.2520 14.85091637 18847
0 PSLV DEB
1 35342U 01049QA  22068.45293639  .00000572  00000-0  29171-3 0  9997
2 35342  98.2923 139.9474 0362047 310.4354  46.5744 13.90366294443094
0 PSLV DEB
1 35344U 01049QC  22066.39708877  .00001827  00000-0  33712-3 0  9995
2 35344  97.8823 111.1759 0051921 247.3757 112.1950 14.65792586467364
0 PSLV DEB
1 35345U 01049QD  22068.70986175  .00004211  00000-0  78587-3 0  9991
2 35345  98.1965  57.2692 0063862  44.1527 316.4748 14.64192957483825
0 PSLV DEB *
1 35346U 01049QE  22068.95292672  .00002547  00000-0  66345-3 0  9990
2 35346  98.8233 150.9183 0039820 209.6044 150.2894 14.50027389 35398
0 PSLV DEB
1 35348U 01049QG  22068.72111061  .00003373  00000-0  55630-3 0  9992
2 35348  98.1250  59.3489 0050619 195.6572 164.3071 14.70292678 53404
0 PSLV DEB
1 35349U 01049QH  22068.32994666  .00027279  00000-0  86501-3 0  9991
2 35349  97.4664 100.9129 0014578 180.4560 179.6671 15.32568406846255
0 PSLV DEB
1 35350U 01049QJ  22068.76869562  .00000911  00000-0  24939-3 0  9998
2 35350  98.6033  64.7516 0074531  99.8340 261.1278 14.48029442457561
0 PSLV DEB *
1 35351U 01049QK  22066.70636923  .00002156  00000-0  63479-3 0  9999
2 35351  98.8179  29.5651 0005211  45.5944 314.5671 14.44732274457505
0 SL-18 DEB
1 35354U 93014BD  22068.76520028  .00021929  00000-0  20751-2 0  9995
2 35354  75.7302 100.7819 0059525 350.7978   9.2117 14.92216400847487
0 SL-18 DEB
1 35356U 93014BF  22068.52236839  .00024275  00000-0  25770-2 0  9998
2 35356  75.7433 162.0127 0062209  57.3622 303.3549 14.87510915466583
0 SL-18 DEB
1 35359U 93014BJ  22068.55187275  .00025514  00000-0  24908-2 0  9992
2 35359  75.7369 156.1582 0054843  50.5279 310.0745 14.91164684775759
0 SL-18 DEB
1 35360U 93014BK  22068.44021735  .00019061  00000-0  20292-2 0  9992
2 35360  75.7343 127.2487 0071107  32.5913 327.9635 14.86997880798827
0 MEASAT 3A
1 35362U 09032A   22068.85021403 -.00000245  00000-0  00000-0 0  9996
2 35362   0.0489 235.0951 0001508  71.5888 258.3869  1.00272659 46607
0 SL-23 R/B
1 35363U 09032B   22068.86809376 -.00000136  00000-0  00000-0 0  9990
2 35363  10.3841 293.2954 4023069 247.0945  71.0076  1.68695634 78305
0 METEOR 2-17 DEB
1 35364U 88005Y   22067.81503681  .00001147  00000-0  84240-3 0  9995
2 35364  82.5500  92.4124 0018834 303.2489 178.0638 13.94853833332534
"""
out_file="""HQ SpOC Sgp4Prop - Version: v9.3 - Build: Jan 19 2024 - Platform: Intel Mac 64-bit - Compiler: gfortran                        


EPHEMERIS GENERATED BY SGP4 USING THE WGS-72 EARTH MODEL
COORDINATE FRAME=TRUE EQUATOR AND MEAN EQUINOX OF EPOCH
USING THE FK5 MEAN OF J2000 TIME AND REFERENCE FRAME

INPUT FILE = ../input/my_test.inp
Start Time =         0.0000 min from epoch
Stop Time  =      1440.0000 min from epoch
Step size  =       360.0000 min


     TSINCE (MIN)           X (KM)           Y (KM)           Z (KM)      XDOT (KM/S)       YDOT(KM/S)    ZDOT (KM/SEC)


 1 34427U 93036RU  22068.94647328 +.00008100  00000 0  11455-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34427  74.0145 306.8269 0033346  13.0723 347.1308 14.7687051569388                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     4194.1215938    -5600.9109006        0.0062409        1.6578326        1.2581947        7.2697321
       360.0000000    -2977.9509927     1002.1867772    -6296.8212519        3.5423948       -6.0821953       -2.6208137
       720.0000000    -1998.4098369     4984.8795254     4532.4581849       -4.2097260        3.2138700       -5.3455282
      1080.0000000     4325.6442690    -4631.3356431     2958.0883019       -0.4030533        3.7980559        6.5286643
      1440.0000000    -1278.7692364    -1542.0567954    -6735.2118438        4.4203485       -6.0572943        0.5737563


 1 34428U 93036RV  22068.90158861 +.00002627  00000 0  63561-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34428  74.0386 139.1157 0038434 196.7068 279.9147 14.5311805268695                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     1222.8426970    -3370.9572866     6108.0935844        5.6811718       -3.7054012       -3.2085877
       360.0000000    -4781.9586367     4928.2514680    -1885.5238169       -2.8545994       -0.1153596        6.9017674
       720.0000000     5175.0828827    -3271.1344175    -3541.8637008       -1.7334307        3.9207385       -6.1769835
      1080.0000000    -2375.1210695     -422.2787377     6676.3139453        5.1382679       -5.2418566        1.4661407
      1440.0000000    -1997.8306504     4002.7245894    -5526.3449371       -5.2650217        3.2308135        4.2109802


 1 34429U 93036RW  22068.93069979 +.00002324  00000 0  48175-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34429  74.0120 358.8898 0044194 129.5211 303.2208 14.6018999069130                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     2179.6982489     1808.1552333     6452.7227411       -7.1678639        0.7612092        2.1730121
       360.0000000     4165.1907935    -1694.5359263    -5505.4783886        6.0463691        1.0501197        4.2512751
       720.0000000    -7049.8136756      236.5474685      -61.7617642       -0.0399077       -2.0724446       -7.2382958
      1080.0000000     4303.3560345     1365.5787278     5424.3044998       -5.9755504        1.5010856        4.3113203
      1440.0000000     2025.1451023    -1978.4363846    -6518.6758652        7.1483484        0.2356190        2.1374960


 1 34430U 93036RX  22068.81563909 +.00000483  00000 0  17604-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34430  74.0532 306.2833 0065617 358.1277   1.9618 14.3459758568223                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     4207.6752763    -5731.5597158       -0.0068613        1.6595938        1.2256652        7.2249223
       360.0000000    -4446.6640314     4448.2344434    -3503.2896247        0.7871419       -4.0863424       -6.1411140
       720.0000000     3350.2378471    -1731.6334400     6049.1231446       -3.0190436        5.9390244        3.4173735
      1080.0000000    -1527.7904366    -1248.2076389    -6880.1423759        4.3112193       -6.0793216        0.1964091
      1440.0000000     -863.8567784     4176.5418086     5772.6589228       -4.4479598        4.5237606       -3.8815514


 1 34431U 93036RY  22068.56203921 +.00000581  00000 0  26231-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34431  74.0773 128.6403 0072810  98.6936 262.2466 14.2283874067563                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -4498.3522072     5626.8682463       -0.0034037       -1.5543957       -1.3207806        7.1512362
       360.0000000     4716.9205731    -4801.3741978    -2586.9985817       -0.1908635        3.3081156       -6.6481278
       720.0000000    -4350.7472660     3545.1365512     4449.6301365        1.7525609       -4.8039612        5.4537213
      1080.0000000     3477.5823092    -1678.5953559    -6127.9180620       -3.1577112        5.7477225       -3.4049797
      1440.0000000    -2226.1535499     -254.6463423     6775.2988403        4.1842855       -6.1192013        1.1271059


 1 34432U 93036RZ  22068.88256954 +.00000495  00000 0  31471-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34432  73.8343 350.6259 0194977 316.7149 104.7023 13.9761246866459                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     3514.2196549     1267.9469789     6286.3509081       -6.2886136        1.9974146        3.2668991
       360.0000000    -4218.5529845     -984.9682907    -5824.1931110        5.9945696       -2.1544325       -3.8024697
       720.0000000     4085.0821116     1020.3417265     5966.9008217       -5.9045320        2.1953923        3.8315854
      1080.0000000    -4750.9577552     -713.9392879    -5455.0186738        5.5418357       -2.3230061       -4.3337098
      1440.0000000     4614.5293752      747.8152916     5597.8407309       -5.4654361        2.3633777        4.3664571


 1 34433U 93036SA  22068.32456284 +.00001354  00000 0  78369-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34433  74.0206  43.9986 0175048 133.2292 228.3613 14.0239876866622                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     5288.6562958     5106.9496790        0.0085139       -1.4723933        1.3781434        7.0381701
       360.0000000    -5071.0533134    -5058.3892054     -595.2800576        1.8049615       -1.1196612       -7.1816515
       720.0000000     5273.2323959     5096.2850252      399.0456862       -1.7563505        1.1107226        7.0283222
      1080.0000000    -5036.5729788    -5039.2254592    -1006.9186080        2.1005279       -0.8384697       -7.1305310
      1440.0000000     5241.0939398     5070.9323178      797.3169875       -2.0443392        0.8473903        6.9955319


 1 34434U 93036SB  22069.18354189 +.00002732  00000 0  20462-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34434  74.0680 145.5195 0335997 236.1481 152.1039 13.7153452065056                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -6013.1125434     2868.9766920     3638.2958745        1.8919755       -3.3755518        5.9989267
       360.0000000     5914.0713483    -4162.4762324      108.7435939        1.1092340        1.7335964       -7.2083973
       720.0000000    -5279.9016584     4606.0193164    -2476.9185105       -3.3139233        0.0746881        6.4997291
      1080.0000000     2621.0585863    -3910.5104985     5758.4872369        5.3716950       -2.3412116       -4.3290453
      1440.0000000     -575.5636927     2770.6120980    -6646.3052515       -6.1608942        3.8216123        1.9137160


 1 34439U 93036SG  22069.19570012 +.00007264  00000 0  12034-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34439  74.0612 359.9013 0119628  27.3462 333.3934 14.6573884069316                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     6980.2351874      -12.0254931       -0.0036709       -0.0453966        2.0854873        7.3069600
       360.0000000    -3721.7032039    -1636.6559645    -5858.3669034        6.3627418       -1.1131658       -3.6748118
       720.0000000    -3429.8854544     1751.5074736     5905.8124463       -6.6120318       -0.8675390       -3.4678799
      1080.0000000     6957.5969656     -327.1269932     -489.7370617        0.5579704        2.0659556        7.2887478
      1440.0000000    -4182.0623526    -1451.6725250    -5592.6540123        6.0388265       -1.3845000       -4.1018521


 1 34442U 93036SK  22068.92772226 +.00002499  00000 0  43362-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34442  74.0170 264.7116 0070210  17.5471  98.7860 14.6718312869525                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     2017.4422618     3046.9496918     6029.0671105       -0.3076302        6.7682355       -3.2496566
       360.0000000     -841.6233099    -6937.2644516     -487.6103977        2.0150568       -0.7282399        7.2600557
       720.0000000    -1170.4047935     4145.8039414    -5635.0395825       -1.8176321       -6.0289011       -4.0178009
      1080.0000000     2094.7714610     2968.7292394     6043.0507595       -0.1270594        6.7857722       -3.2221131
      1440.0000000    -1023.6353550    -6910.5133984     -512.8569899        1.9918067       -0.8083633        7.2584346


 1 34445U 93036SN  22068.73338576 +.00009736  00000 0  88849-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34445  74.0124  41.2799 0085291 113.2816 310.9216 14.9238949970143                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     1203.1040533     3326.9180148     5951.8440213       -5.7402455       -3.8103689        3.2258110
       360.0000000     5121.2182369     2992.7547483    -3761.6614808        2.0170570        4.0126171        6.0114841
       720.0000000    -2450.8239137    -4023.5025839    -5180.9646035        5.2522151        2.7008778       -4.6614027
      1080.0000000    -4479.4815758    -1898.7462069     4905.0201625       -3.4443039       -4.6998466       -4.9206612
      1440.0000000     3817.5212740     4488.2039997     3691.3108277       -4.3818564       -1.3500167        6.0367564


 1 34446U 93036SP  22068.81213567 +.00000225  00000 0  68441-4 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34446  74.0404 108.2372 0054093 229.4135 130.2310 14.4703272668807                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -2234.8160211     6782.3921086       -0.0054909       -1.9542043       -0.6194979        7.1726623
       360.0000000     2848.1129910    -4660.3614033    -4507.3845379       -0.0456900        5.2145495       -5.4163816
       720.0000000    -2123.6186164      265.1531472     6809.4177051        1.9649719       -7.1385878        0.8649564
      1080.0000000      324.7230138     4421.2480891    -5560.9621379       -2.9067645        5.5035648        4.1572491
      1440.0000000     1503.8239441    -6746.5911116     1602.3143099        2.4152965       -1.1014242       -7.0224045


 1 34447U 93036SQ  22068.71718351 +.00027570  00000 0  27166-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34447  74.0222 176.1774 0021532 277.1236 193.9740 14.9175748769753                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     2384.4543115    -1956.8956639     6259.8454773        7.0779080        0.2752340       -2.6089011
       360.0000000    -6801.3946747       72.8154217     1536.5039108        1.5687086       -2.1497362        7.0783213
       720.0000000     -460.3106188     1955.6870664    -6666.2689495       -7.5233113        0.4481770        0.6513677
      1080.0000000     6926.6071083     -744.6283712      342.8053559        0.5698102        2.0331152       -7.2597506
      1440.0000000    -1480.6349073    -1747.1402276     6591.1398929        7.3392959       -1.1373845        1.3490718


 1 34449U 93036SS  22069.15291990 +.00003563  00000 0  93156-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34449  74.0534 185.4595 0067494 230.9988 198.3222 14.4839444268598                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -2370.5137270    -2069.9766877     6417.9695296        6.9947597       -0.0604554        2.5413555
       360.0000000    -2915.1955791     1547.4897052    -6278.8942371       -6.8265539       -1.3921348        2.7765918
       720.0000000     6462.0459725     -324.1032770     2912.7807847        3.0106328        2.1118076       -6.5408333
      1080.0000000    -6711.6803296    -1107.4683810     2202.9727338        2.5015947       -1.7658573        6.7883137
      1440.0000000     3242.0513719     1917.0968744    -5981.9438446       -6.6672294        0.5677436       -3.4459877


 1 34451U 93036SU  22068.79243827 +.00003868  00000 0  10233-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34451  74.0627 200.4329 0067663 231.7496 127.7563 14.4779143268539                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -6693.8585677    -2493.8002526        0.0010395        0.6841345       -1.9286277        7.1699397
       360.0000000     4498.0817925     3015.7765170    -4544.5128050       -5.2596855       -0.2769898       -5.3811606
       720.0000000      -48.4027335    -2088.0294518     6831.6000184        7.0527713        2.2805300        0.7163993
      1080.0000000    -4606.9129494       46.4709914    -5411.2291330       -5.2277827       -3.1199029        4.3612235
      1440.0000000     6713.0150879     1858.4984391     1314.2784226        0.6955720        2.3698416       -7.0961194


 1 34452U 93036SV  22068.79425045 +.00001866  00000 0  22273-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34452  74.0450  58.8413 0116185 144.3611 286.0955 14.8028215770176                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     -248.7478663     3054.9441884     6269.1252905       -4.2892725       -5.7073579        2.5112661
       360.0000000     3957.5547912     4135.5335443    -4190.3675860        1.0268722        4.7427570        5.6513085
       720.0000000    -2098.6133809    -5422.4588265    -3872.9914589        3.7320089        2.7225159       -5.9903692
      1080.0000000    -2506.2165041     -493.2071567     6447.4746277       -3.5038674       -6.4844212       -1.9121937
      1440.0000000     3962.8362090     5847.0819359     -430.6580622       -1.4885244        1.4685328        7.1722534


 1 34453U 93036SW  22068.88937838 +.00002522  00000 0  63547-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34453  74.1682 233.2845 0047826 119.5810   3.3000 14.5095976768649                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     3593.7523505     2113.6777963     5699.0910241        2.8860292        5.7353488       -3.9358248
       360.0000000    -4538.7186366    -5296.5541808    -1463.0700071        0.6427941       -2.4470835        7.0258045
       720.0000000     2729.0540723     5322.7781439    -3842.6730846       -3.8894243       -2.2686682       -5.9771577
      1080.0000000      617.4311501    -2275.7471826     6662.1924225        4.8692933        5.5552569        1.4193231
      1440.0000000    -3756.5273908    -2119.3778986    -5691.2763639       -2.9554391       -5.5303379        4.0171876


 1 34454U 93036SX  22068.91971155 +.00000319  00000 0  11812-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34454  74.0583 280.7094 0037596 327.9100  31.9764 14.3584487368332                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     1324.9963533    -7006.0380745        0.0042097        2.0216896        0.3748169        7.2028616
       360.0000000    -2097.7529620     5841.6932446    -3584.7870291       -1.0226104       -4.1410331       -6.1068387
       720.0000000     2273.7484677    -2815.7281497     6163.3800513       -0.2385918        6.7432547        3.1943934
      1080.0000000    -1833.6275976     -901.8432877    -6841.9624736        1.3812602       -7.3160884        0.6172601
      1440.0000000      887.2548160     4511.5704522     5502.7444000       -2.1307407        5.6890571       -4.2993633


 1 34456U 93036SZ  22068.85349922 +.00005488  00000 0  13805-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34456  74.0462 226.5559 0056524 251.9784 107.5213 14.5047791768504                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -4894.0807149    -5167.3755252        0.0073425        1.4691576       -1.4364652        7.1913166
       360.0000000     2540.4975040     4593.7164981    -4739.9946884       -4.7096820       -2.7605009       -5.1742212
       720.0000000     1323.0840678    -1452.4634624     6856.9781451        5.2332622        5.3063023        0.1002437
      1080.0000000    -4500.4512212    -2558.2263119    -4845.9496455       -2.7560701       -4.8156309        5.0493271
      1440.0000000     5074.2971670     4951.2725003      165.0258775       -1.3432972        1.5708211       -7.2158442


 1 34457U 93036TA  22068.86184197 +.00003162  00000 0  55362-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34457  74.0015 304.7929 0101048  52.3409  79.2774 14.6494571369328                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -1561.5009731     4742.0218434     4961.5926331       -4.3286689        3.7873908       -4.8640565
       360.0000000     4142.9754079    -5556.8135703     1009.8891442        1.0275028        2.1489374        7.1896882
       720.0000000    -3046.0359722     1289.5151601    -6317.1400939        3.1620542       -6.1358472       -2.7545753
      1080.0000000     -999.4665491     4353.4912321     5431.3408336       -4.3736803        4.4305190       -4.2510417
      1440.0000000     3865.7208327    -5840.8235090      269.4444802        1.5499499        1.4381883        7.2675252


 1 34458U 93036TB  22068.41282814 +.00004795  00000 0  11759-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34458  74.0292 153.1192 0016214 199.0657 287.1028 14.5260498468710                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     3007.0442588    -3292.7944229     5508.0749596        5.9564679       -1.6572325       -4.2480382
       360.0000000    -6156.0339221     3455.1733468     -839.5568848       -1.7535479       -1.3923271        7.1409233
       720.0000000     5391.8916102    -1428.0288842    -4377.8703204       -3.5162342        3.6446053       -5.5314804
      1080.0000000    -1312.4720147    -1506.6228419     6804.6802266        6.5263026       -3.6551943        0.4367740
      1440.0000000    -3606.1007217     3598.5402469    -4955.0398207       -5.4424724        1.3740485        4.9462950


 1 34459U 93036TC  22069.08344769 +.00000282  00000 0  18354-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34459  74.0784 101.5573 0196934 142.0480 219.4723 13.9752012466513                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -1481.5557486     7244.9805915        0.0028255       -1.9377053       -0.4952242        7.0071046
       360.0000000     1394.0419689    -7033.1117888      -36.7722076        2.0329650        0.3403053       -7.2273418
       720.0000000    -1183.5688329     7277.1760619     -658.0527724       -2.0635225        0.2063983        6.9780204
      1080.0000000     1093.1661903    -7050.3031663      634.2945833        2.1579085       -0.4038165       -7.1956004
      1440.0000000     -875.6586307     7238.9359445    -1310.5404659       -2.1504618        0.9075864        6.8882842


 1 34464U 93036TH  22068.92676281 +.00002223  00000 0  40696-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34464  74.0215 280.4680 0053730  59.9090  56.2251 14.6554125069477                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     1126.8066194     3418.2674393     6034.6481069       -2.1344966        6.4759781       -3.2234509
       360.0000000     1011.1588794    -6930.6759598     -719.1231614        2.1622190       -0.3878856        7.2067026
       720.0000000    -2245.9126935     3939.0047991    -5460.5877032       -0.2298023       -6.0947028       -4.3092298
      1080.0000000     1359.0081650     2880.3154260     6260.9019042       -1.8575791        6.8013090       -2.6835539
      1440.0000000      669.0036823    -6896.1436926    -1265.9112678        2.2177491       -1.0470828        7.1208317


 1 34465U 93036TJ  22068.68041544 +.00002908  00000 0  64829-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34465  74.0759 152.8527 0089581 164.8833 195.5040 14.5484984668736                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -6363.9852719     3263.2513159       -0.0100655       -0.9078413       -1.8255371        7.1500641
       360.0000000     4697.6581238     -774.6843142    -5223.7805454       -4.4665457        3.8403023       -4.6717192
       720.0000000     -133.8168187    -2107.1119803     6731.7212017        6.7116394       -3.2738921       -0.9542481
      1080.0000000    -4762.9861874     3808.9147127    -3740.9923069       -4.3757061        0.4380696        5.9881644
      1440.0000000     6169.2507614    -2801.0064776    -1909.5849976       -0.8611445        2.7543584       -6.9803018


 1 34468U 93036TM  22068.66449296 +.00000505  00000 0  15296-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34468  74.0161 160.4309 0082400 276.6917  82.4866 14.4307695868614                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -6710.5283700     2385.4364412       -0.0062517       -0.7406719       -1.9233533        7.1975725
       360.0000000     5749.9210525     -851.5383513    -4081.1166468       -3.6874961        3.0980614       -5.7549090
       720.0000000    -2301.7854145    -1186.9324097     6690.9842450        6.5662797       -2.9888700        1.7474648
      1080.0000000    -2069.3137377     2698.1473682    -6207.4484868       -6.7207639        1.6720573        2.9452399
      1440.0000000     5508.4891578    -3173.9014705     3286.8722512        3.9375204        0.3950653       -6.3098988


 1 34469U 93036TN  22068.86934120 +.00007221  00000 0  12494-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34469  74.0174 358.6480 0010578 107.0254 253.2076 14.6847699669202                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     7046.9185053     -166.3154637        0.0017290        0.0335397        2.0692736        7.2302625
       360.0000000    -3423.7639546    -1599.0135279    -5958.9429333        6.5557128       -1.1989385       -3.4528637
       720.0000000    -3708.3776180     1785.5028612     5700.9639780       -6.3979563       -0.8532051       -3.8832001
      1080.0000000     7033.3830228     -233.8681102      385.3139590       -0.3419596        2.0854145        7.2187458
      1440.0000000    -3109.1010032    -1582.3172603    -6133.8506335        6.7218481       -1.2709209       -3.0862200


 1 34472U 93036TR  22069.05485062 +.00001254  00000 0  10568-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34472  74.0854 185.8782 0344216 270.6029  85.5662 13.6488918765000                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -7348.8541959     -756.6040178       -0.0011646       -0.0381723       -2.0284147        7.0670870
       360.0000000     5969.8320215     -735.3235913     4560.5231044        4.2010799        1.9746549       -5.4922938
       720.0000000    -3012.8746808     1543.0240017    -6319.2816230       -6.8778528       -1.4156650        2.8183286
      1080.0000000    -1079.8307324    -2162.6733691     7247.9251210        6.9992565        0.2405809        1.1524547
      1440.0000000     4952.4168267     1779.4939419    -4933.8634275       -5.5601980        1.0062705       -4.9674333


 1 34473U 93036TS  22068.77072253 +.00004133  00000 0  10903-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34473  74.0323 191.9444 0023667 209.7653 274.0908 14.4910411368650                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     4185.3433656     -777.6652051     5682.5285472        5.8603016        2.4088101       -3.9999967
       360.0000000    -6901.5357257     -970.9177970    -1472.3570156       -1.1810972       -2.2872839        7.0172776
       720.0000000     5757.4067177     2153.6111466    -3544.0305386       -4.1585989        0.9870822       -6.1651765
      1080.0000000    -1508.0128153    -2200.3635440     6592.0193503        7.1931688        0.7881742        1.8909254
      1440.0000000    -3667.3410853     1092.6392685    -5996.8132935       -6.2920308       -2.1170878        3.4424190


 1 34474U 93036TT  22069.06593345 +.00001669  00000 0  10117-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34474  74.0826 178.2287 0229473 223.6924 311.6293 13.9490386366320                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     7124.8492804     -445.9400469      790.2107752        0.7999182        2.0239831       -7.1848321
       360.0000000    -7311.8511224      556.5417467     -965.5770411       -1.1888151       -1.9305004        6.9285875
       720.0000000     6855.4326282     -892.1590120     2021.9307535        2.1411097        1.8769149       -6.9228568
      1080.0000000    -6982.7306940      989.4092694    -2158.0893449       -2.4567416       -1.7681382        6.6544675
      1440.0000000     6352.6026201    -1298.2092783     3188.4252071        3.3956467        1.6322439       -6.4401889


 1 34476U 93036TV  22068.63364977 +.00001641  00000 0  12974-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34476  74.0120 200.3933 0366042 353.1964   6.4250 13.6516465164789                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -6681.9346657    -2484.0977131       -0.0069256        0.7072201       -1.9732639        7.3199250
       360.0000000     6557.1220981     1250.5482588     3717.1696337        2.8997079        2.8317173       -5.8329724
       720.0000000    -3222.8177404      817.3596935    -6452.4517705       -6.1432128       -3.0783105        2.9410689
      1080.0000000      469.6552625    -1997.4706019     7118.9466224        6.9454749        2.2750390        0.4578313
      1440.0000000     4193.7841360     3076.7934481    -5458.7124541       -5.6743345       -0.6001774       -4.3871361


 1 34477U 93036TW  22069.22728087 +.00001096  00000 0  30407-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34477  73.9823 137.9328 0092528 285.4009 128.2922 14.4532807568747                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -4164.7217133     1593.4273477     5595.7524221        3.6689192       -4.9324096        4.1946914
       360.0000000      974.7474368     1745.9093136    -6772.3440770       -5.6363151        4.9813734        0.4869955
       720.0000000     2654.4217843    -4430.3226496     4976.7564821        4.8773237       -2.6040233       -4.9566806
      1080.0000000    -5028.4968548     4983.7635509     -548.2907729       -1.8982500       -1.0411811        7.1858346
      1440.0000000     5003.5722793    -3303.1516297    -3805.3832687       -1.8635166        4.1673203       -5.9531396


 1 34478U 93036TX  22068.38240896 +.00002447  00000 0  57864-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34478  74.0630 146.6712 0096906 357.3956 122.6783 14.5184435868786                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     2151.2372242    -3423.0411202     5872.6721975        5.9373241       -2.6598559       -3.6438807
       360.0000000    -5406.5111922     4186.6936053    -1650.4554827       -2.6382977       -0.6577364        7.0582479
       720.0000000     5612.5799168    -2570.4567160    -3625.8552871       -2.2883826        3.6496398       -6.0537875
      1080.0000000    -2147.0623606     -839.7368385     6696.6843718        5.8300346       -4.5206315        1.3760879
      1440.0000000    -2265.7721676     3575.4101545    -5651.8341698       -5.7624769        2.6487315        4.0586301


 1 34481U 93036UA  22068.93573963 +.00000769  00000 0  32254-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34481  73.9822  34.7473 0076080 159.1313 262.1452 14.2604377667828                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     1936.8069738     3441.9765477     6002.4017756       -5.9111794       -2.8831118        3.4861416
       360.0000000      603.6073955    -1987.3204947    -6899.3003984        6.2117512        4.0048115       -0.6670416
       720.0000000    -2736.3520887      441.8477144     6582.2606909       -5.6209773       -4.4871947       -2.0797037
      1080.0000000     4753.6215166     1310.0397070    -5298.7799503        3.9890942        4.2047772        4.5854789
      1440.0000000    -5816.6680385    -2704.8959110     3106.3020970       -1.8582917       -3.4020566       -6.4306469


 1 34482U 93036UB  22068.94395123 +.00001626  00000 0  50637-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34482  74.0495 285.7402 0073443 330.2917 141.2161 14.4019552068264                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     1027.7881630     3088.9471135     6388.9122546       -2.6014216        6.4388484       -2.6509633
       360.0000000      543.4400792    -6016.3292070    -3710.3681296        2.7507473       -3.4886767        6.0716294
       720.0000000    -1917.3528195     6915.6714473     -297.2894945       -1.8825646       -0.8523431       -7.1350078
      1080.0000000     2564.8596830    -5037.0908440     4321.3485695        0.3875247        4.9521201        5.6130670
      1440.0000000    -2313.6283285     1532.1883807    -6566.5845902        1.1657568       -7.1055630       -2.0117166


 1 34486U 97051GB  22068.70192972 +.00002799  00000 0  67048-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34486  86.4080 256.1055 0019315 167.8443 327.5296 14.5310581668840                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     1511.7369097     4812.1012529     4966.1058682        0.9434253        5.2063182       -5.3304813
       360.0000000    -1702.7740731    -6897.6119304      266.2300945        0.5250169        0.1706904        7.4620454
       720.0000000      793.3192484     4544.9224886    -5389.3931440       -1.6837094       -5.4475052       -4.8613034
      1080.0000000      611.1354704      614.6088753     7028.1327364        1.7833178        7.2435058       -0.8006316
      1440.0000000    -1667.6353938    -5450.7000314    -4254.6019652       -0.7459349       -4.4353689        5.9711856


 1 34487U 97051GC  22068.94053602 +.00000335  00000 0  99357-4 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34487  86.3423 241.7417 0031623 345.4457 146.4254 14.4014241068458                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     2569.8533796     4066.1820715     5291.1596386        2.3441670        5.0327827       -4.9771929
       360.0000000    -3385.5392135    -6056.5000187    -1589.2060357       -0.3843428       -1.6923403        7.2916483
       720.0000000     3003.3166587     5902.6215240    -2719.2406630       -1.7422829       -2.2915776       -6.8728687
      1080.0000000    -1463.1694379    -3492.9076117     6035.0546934        3.2561440        5.4482400        3.9613843
      1440.0000000     -563.2953317      -81.5875397    -7112.8815858       -3.5588251       -6.5543995        0.3792841


 1 34488U 97051GD  22068.64888950 +.00003454  00000 0  76260-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34488  86.3495 209.5295 0026344 115.9819 244.4107 14.5686251168974                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -6171.3220790    -3495.7614091       -0.0052928        0.2573569       -0.4022323        7.4787560
       360.0000000     3710.7867876     2496.4418964    -5509.9191633       -5.2083256       -2.5925291       -4.7074624
       720.0000000     1390.8880233      274.7780093     6912.1585988        6.3943804        3.6947669       -1.4354129
      1080.0000000    -5533.9832856    -2842.2961861    -3425.9100478       -2.9407910       -2.1215380        6.5441158
      1440.0000000     5653.1323329     3340.0072310    -2665.9325238       -2.6652408       -0.9764435       -6.9409319


 1 34489U 97051GE  22068.70986860 +.00005650  00000 0  16117-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34489  86.4090 318.3848 0074197 128.7816 232.0044 14.4344221667989                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     5353.8280318    -4755.8920051       -0.0065524        0.2712757        0.3823330        7.4303321
       360.0000000    -4269.5820354     3421.8005848    -4576.6421971        3.3110407       -3.4351905       -5.7481245
       720.0000000     1506.8004404     -768.7495464     6879.3338367       -5.3871320        4.9631496        1.6871268
      1080.0000000     2047.2079768    -2384.3149129    -6460.7595728        5.1060368       -4.3167663        3.2006527
      1440.0000000    -4612.3470743     4416.6437734     3052.7005055       -2.7252830        1.8861312       -6.7620080


 1 34490U 97051GF  22069.05151916 +.00096914  00000 0  60495-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34490  86.3078 166.2021 0026908 104.4790  16.6375 15.0875766969499                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     3384.3667474    -1222.2197562     5881.3184604        6.3861376       -1.3073117       -3.9310127
       360.0000000    -5368.4087307     1049.7366397     4218.3479891        4.4442140       -1.4987783        5.9821673
       720.0000000    -4673.5743099     1495.3322628    -4908.2519803       -5.2748777        0.9643456        5.3351308
      1080.0000000     4179.5504550     -688.5868752    -5493.4152606       -5.7394391        1.7544520       -4.6082050
      1440.0000000     5718.3054414    -1687.3899422     3480.8890985        3.8498362       -0.5448260       -6.5379999


 1 34492U 97051GH  22068.21456637 +.00003667  00000 0  20620-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34492  86.4992 148.4187 0300676 196.1027 197.7228 13.8987865565777                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -5516.2078935     3099.7518085     4054.7464170        3.1659352       -2.3760722        5.9888208
       360.0000000     5585.1050055    -3234.0669849    -2948.8471020       -2.4411619        2.0032332       -6.9175643
       720.0000000    -6268.2611884     3751.0319374     1818.8914208        1.2572089       -1.2777269        6.9405899
      1080.0000000     6027.1183217    -3713.0752480     -462.7263629       -0.2046123        0.6724798       -7.5748384
      1440.0000000    -6338.5145469     3998.7739408     -613.5972979       -0.7800083       -0.0274854        7.1344724


 1 34493U 97051GJ  22068.71924692 +.00015517  00000 0  23522-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34493  86.2631 163.3558 0057545 248.7354 110.7712 14.7291424368939                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -6752.2932596     2018.6193195       -0.0011897       -0.1712196       -0.4596390        7.4984203
       360.0000000     3025.7629441     -481.8418417    -6290.3139869       -6.4616295        2.1689532       -3.2722424
       720.0000000     4094.0598096    -1617.5189074     5499.6016053        5.6949414       -1.4075134       -4.6916854
      1080.0000000    -6574.3637834     1890.6239803     1728.4374531        1.6004431       -0.9848954        7.2659357
      1440.0000000     1434.7343925       26.8334539    -6849.1920960       -7.0549133        2.2721476       -1.4795395


 1 34496U 97051GM  22068.11986750 +.00008875  00000 0  22227-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34496  86.3807 313.2362 0079654 214.8210 144.7768 14.4941140550752                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     4901.3494152    -5212.8037544       -0.0074885        0.3599405        0.3024596        7.4261364
       360.0000000    -3724.5359320     3527.5157092    -4845.6009990        3.2883905       -4.0162401       -5.4692535
       720.0000000      656.5798233      -44.7589278     7101.5911127       -5.0563129        5.4602122        0.4530754
      1080.0000000     2954.9959113    -3674.5096473    -5336.3293472        4.0712431       -3.9259168        4.8817682
      1440.0000000    -4734.9452913     5183.9588653      770.1396919       -0.8876526        0.2624721       -7.4809396


 1 34497U 97051GN  22069.07755535 +.00000714  00000 0  30910-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34497  86.2760 274.9045 0091436  57.6087  72.1578 14.2066468467474                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000      -49.6613096     4706.4615099     5418.4618440       -0.7987482        5.6457884       -4.8141236
       360.0000000      244.0934794    -6004.7436560    -3978.3528697        0.7421345       -4.0043502        6.2191716
       720.0000000     -465.2098206     7009.5632392     1687.0507085       -0.6132623        1.7622449       -7.1798106
      1080.0000000      590.5406155    -7134.0902781      277.0499500        0.4553723        0.3898679        7.4577149
      1440.0000000     -701.0636480     6714.7318392    -2674.6576995       -0.2349970       -2.7115044       -6.8592293


 1 34503U 97051GU  22068.79645465 +.00004812  00000 0  97823-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34503  86.3205 182.9085 0013065 242.5442 117.4440 14.6089416069130                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -7066.6648243     -359.0325042       -0.0054580        0.0236437       -0.4807339        7.4894947
       360.0000000     4130.6313607      569.3713139    -5705.0202765       -6.0809340       -0.0137109       -4.4044348
       720.0000000     2181.3140818     -330.2474267     6714.7420986        7.1285314        0.4820453       -2.2949987
      1080.0000000    -6724.3632343     -160.9343822    -2195.1409513       -2.3139441       -0.5618468        7.1165548
      1440.0000000     5709.7801092      511.7068539    -4131.5084192       -4.4141155        0.2008723       -6.0744697


 1 34508U 97051GZ  22069.15122779 +.00001244  00000 0  50236-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34508  86.3768 317.1334 0077286 309.0603  50.3727 14.2557850167667                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     5242.3927837    -4865.8405943       -0.0049641        0.3486763        0.3213129        7.4699822
       360.0000000    -5036.2406999     4469.1224865    -2567.7784031        1.6766940       -2.1592280       -6.9067153
       720.0000000     3952.2580269    -3257.6660166     5042.6935517       -3.5675023        3.7960011        5.3174002
      1080.0000000    -2459.8578099     1743.3866810    -6499.7882598        4.8301963       -4.7978185       -3.0601215
      1440.0000000      419.7768290      231.8146067     7205.4588147       -5.3805405        5.0830355        0.1843773


 1 34511U 97051HC  22068.81147505 +.00005535  00000 0  13651-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34511  86.3845 264.6079 0030984 200.9344 169.6160 14.5172389868557                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     -576.4904320    -6978.9173280     1294.9499160        0.5892750        1.3165247        7.3315965
       360.0000000        3.0472266     3924.8720182    -5901.6202783       -0.8592447       -6.1964279       -4.1395574
       720.0000000      576.2900489     1382.7713292     6933.8179584        0.6205179        7.3080775       -1.5295798
      1080.0000000     -838.0487433    -5973.9879659    -3778.7548131       -0.0007354       -4.0068359        6.3079921
      1440.0000000      594.6877987     6859.8489212    -1651.3944139       -0.6358850       -1.6959122       -7.2932044


 1 34517U 97051HJ  22068.85833177 +.00022186  00000 0  23199-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34517  86.3620 190.4031 0002835  70.8380  53.2461 14.8954832669403                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     3917.3827477      346.8629021     5755.8724083        6.1098545        1.3950249       -4.2309426
       360.0000000    -6254.8399973    -1315.9885780     2793.8461286        3.0746795        0.1120233        6.9077012
       720.0000000    -1676.7917688      137.4389781    -6780.8760728       -7.1997156       -1.4050756        1.7534874
      1080.0000000     6857.8892937     1242.9372545     -411.9748557       -0.5140370        0.3946726       -7.5290568
      1440.0000000     -764.1944272     -579.7611436     6901.7226912        7.4024269        1.2380459        0.9223850


 1 34518U 97051HK  22068.87864802 +.00040197  00000 0  38222-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34518  86.3834 211.8107 0021328 221.9151 138.0446 14.9310318169350                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -5932.0764928    -3679.5774542        0.0043505        0.2486788       -0.4059670        7.5372373
       360.0000000      488.1121402      814.1574964    -6897.7217142       -6.4159329       -3.8932273       -0.9246073
       720.0000000     5801.4763526     3439.0460003     1714.6073562        1.3479223        1.3718581       -7.3270967
      1080.0000000    -1959.0010339    -1679.1691897     6477.6631741        6.0906666        3.5242679        2.7413382
      1440.0000000    -5360.2979771    -3021.7969547    -3290.7150775       -2.8375464       -2.2195949        6.6380677


 1 34520U 97051HM  22068.88279414 +.00067561  00000 0  46845-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34520  86.3574 191.0984 0004408 115.1070 245.0623 15.0529876969422                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -6803.5015308    -1334.5968454        0.0033004        0.1036766       -0.4702996        7.5680850
       360.0000000     -564.6994601      338.6932842    -6907.6317102       -7.4032361       -1.4708604        0.5320302
       720.0000000     6727.7817328     1354.2687196     -976.4542681       -1.1311653        0.2684539       -7.4948783
      1080.0000000     1480.2245097     -157.2225135     6756.1896826        7.2787211        1.4819382       -1.5576984
      1440.0000000    -6535.6816086    -1347.1765186     1867.7068866        2.1115491       -0.0754815        7.2862234


 1 34521U 97051HN  22068.82096123 +.00001811  00000 0  52005-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34521  86.3823 263.4130 0006431 189.2578 170.8506 14.4439323368538                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     -817.8655603    -7082.6804868        0.0008164        0.4691413       -0.0470551        7.4614626
       360.0000000      358.5926859     5503.6235864    -4510.4444362       -0.9146439       -4.6643244       -5.7725751
       720.0000000      265.5077098    -1488.9512983     6953.8969729        0.9622795        7.2633531        1.5108708
      1080.0000000     -792.4036207    -3220.2080330    -6313.0127800       -0.5811637       -6.6036125        3.4373988
      1440.0000000      974.6906598     6465.8446894     2811.6946243       -0.0674762        3.0009231       -6.8593664


 1 34522U 97051HP  22068.93100740 +.00024441  00000 0  40753-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34522  86.3626 249.0936 0048383 205.8879 303.0979 14.6908161968640                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     2356.0524420     5517.2862465     3649.2800065        1.0136246        3.7999661       -6.4396486
       360.0000000    -1983.4701381    -5791.2717425     3546.3964506        1.7362270        3.3751570        6.4566736
       720.0000000     -511.0039934      -83.9070488    -7017.5224990       -2.6956980       -7.0187836        0.2464345
      1080.0000000     2449.5165360     5736.5885692     3219.1971660        0.8542183        3.3723934       -6.6995636
      1440.0000000    -1897.8462626    -5546.1197346     3958.5023628        1.9001776        3.7822883        6.1797671


 1 34524U 97051HR  22069.18279791 +.00007169  00000 0  16785-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34524  86.3435 235.2795 0039177 195.3316 164.6708 14.5400745368652                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -4055.7110177    -5852.7162385        0.0088557        0.3913587       -0.2710289        7.4541826
       360.0000000     2422.3789215     4069.8361375    -5255.1218329       -3.4500611       -4.3971024       -5.0202751
       720.0000000      730.4036794      258.8756634     7044.8162984        4.2375867        6.1428146       -0.6930186
      1080.0000000    -3507.5674206    -4530.4579789    -4218.8081081       -2.2451000       -3.8662261        5.9863321
      1440.0000000     3920.4756008     5718.3525973    -1370.1230206       -1.2230727       -0.9193362       -7.3681359


 1 34525U 97051HS  22068.83195944 +.00000608  00000 0  19645-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34525  86.3704 269.8024 0002882 299.3778  72.1459 14.3750870368319                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000       65.6275410    -7004.0011667     1414.4644682        0.4683181        1.4869707        7.3066827
       360.0000000     -290.6288665     5100.2479153    -5004.1211467       -0.3642632       -5.2295296       -5.3092379
       720.0000000      430.2207104    -1560.7544568     6953.3092077        0.1550543        7.2885506        1.6246867
      1080.0000000     -447.1103815    -2475.4789422    -6693.0846993        0.1007827       -6.9964772        2.5818835
      1440.0000000      332.2267873     5728.7018958     4255.7485328       -0.3304292        4.4691045       -5.9760817


 1 34526U 97051HT  22068.91198510 +.00026319  00000 0  30963-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34526  86.3397 196.8887 0017449 225.1990 134.7816 14.8461487869181                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -6702.5561896    -2034.9486283        0.0027821        0.1385307       -0.4606996        7.5252535
       360.0000000     1563.0261555      922.4290768    -6750.1111974       -7.0258377       -1.9904189       -1.9063026
       720.0000000     5887.1636803     1531.8117563     3438.9575232        3.4473125        1.4693639       -6.5614082
      1080.0000000    -4591.6647031    -1698.1348109     5004.9165466        5.2789042        1.2184149        5.2427928
      1440.0000000    -3593.2166846     -661.1648249    -5968.2925548       -6.1166607       -2.0632903        3.8973013


 1 34529U 97051HW  22068.77403985 +.00000428  00000 0  13464-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34529  86.3611 259.3456 0007967  48.6875 311.5008 14.3808390368380                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -1320.4928427    -7019.1293618        0.0081037        0.4681615       -0.0755685        7.4590690
       360.0000000      862.2883898     5890.0789852    -3969.1164504       -1.1597496       -3.9953524       -6.1899314
       720.0000000     -113.7303280    -2805.2047876     6555.2994461        1.4798718        6.7305378        2.9022710
      1080.0000000     -685.7272926    -1189.4540440    -7020.7436020       -1.3103491       -7.2105452        1.3547908
      1440.0000000     1275.1328229     4819.8567743     5109.5775810        0.7083131        5.3326543       -5.1866393


 1 34532U 97051HZ  22068.93917639 +.00029831  00000 0  40418-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34532  86.3340 225.2890 0034538  76.4842 284.0222 14.7838122968744                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -4931.6812161    -4981.6842022       -0.0051653        0.3660977       -0.3154984        7.5300562
       360.0000000     1375.9766117     1986.1824759    -6616.6429276       -5.0906777       -4.8920303       -2.5303209
       720.0000000     4007.3245203     3610.9870664     4460.3014403        3.1626304        3.6932172       -5.7793164
      1080.0000000    -4167.0986207    -4467.7324260     3404.2430066        2.9246009        2.3245532        6.5728220
      1440.0000000    -1102.5692599     -468.8178840    -6938.8780165       -5.1978964       -5.2754312        1.1942101


 1 34535U 97051JC  22068.72880154 +.00021552  00000 0  22690-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34535  86.3316 164.9437 0023416  87.7299 272.6608 14.8907248369565                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -6741.7127201     1813.5383466        0.0025535       -0.1006438       -0.4733666        7.5427227
       360.0000000     1320.5939508       98.9752712    -6876.2289154       -7.1237587        2.0208545       -1.3415266
       720.0000000     6252.7503622    -1872.6265271     2456.9752756        2.7192424       -0.2748097       -7.0536473
      1080.0000000    -3647.6271792      612.7793223     5894.4470180        6.1388279       -1.9562179        3.9840805
      1440.0000000    -4868.8451893     1667.6252328    -4737.9693501       -5.0127417        1.0239775        5.5362660


 1 34538U 97051JF  22068.71389750 +.00001304  00000 0  47311-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34538  86.2789 240.6891 0045457 130.9806   1.3945 14.3205024368032                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     2652.9439101     4027.9806538     5246.1401757        2.4244371        4.9879172       -5.0430938
       360.0000000    -3461.2875164    -5813.2512584    -2452.3997408       -0.8407860       -2.4137720        6.9733642
       720.0000000     3417.1326298     6180.9667428    -1117.6000449       -0.9765218       -0.7517356       -7.3734003
      1080.0000000    -2632.9009438    -5179.5720280     4187.3122958        2.5180395        3.6367874        6.0079734
      1440.0000000     1137.2864761     2843.2554934    -6499.8680781       -3.5140635       -5.7367471       -3.1585471


 1 34540U 97051JH  22069.09230438 +.00001930  00000 0  45493-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34540  86.3285 200.0620 0028056 167.6445   5.2521 14.5356914769000                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     6615.1240135     2357.3762018      854.9420854        0.7013319        0.7645020       -7.4468253
       360.0000000    -5019.2431023    -2132.7758813     4547.8945549        4.6601842        1.3020216        5.7130091
       720.0000000       83.4521178      513.7264615    -7084.5947856       -7.0352473       -2.5191996       -0.2862279
      1080.0000000     4809.1233654     1384.1992601     4995.7967841        4.8936085        2.1163321       -5.2995795
      1440.0000000    -6690.1232157    -2403.5335796      268.6369611        0.4408110       -0.3504354        7.4559886


 1 34542U 93036UC  22069.12781667 +.00004135  00000 0  70743-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34542  74.0223 311.9362 0024129  80.2712 280.1181 14.6898757169218                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     4706.2809753    -5238.5780915       -0.0105218        1.5232125        1.4028092        7.2364948
       360.0000000    -3479.1294507     1341.6936385    -6001.2510490        3.6569664       -5.6013936       -3.3800742
       720.0000000    -1342.3605004     4000.3786423     5616.4978977       -4.9840958        3.9905671       -4.0086759
      1080.0000000     4674.7464708    -5234.9308061      547.5069937        1.1577528        1.8227101        7.2132069
      1440.0000000    -3167.3340993      994.4286070    -6237.4606987        3.7959829       -5.7954906       -2.8565069


 1 34544U 93036UE  22068.94281865 +.00000383  00000 0  37993-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34544  73.9154 262.0212 0393317 184.2596 297.2850 13.5182939164300                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     2249.4171994     3094.0319597     6232.4654592       -0.1056486        6.5575946       -3.5214715
       360.0000000    -1858.0412437     1293.2997713    -7024.5944729       -1.4111690       -7.1325157       -1.2365801
       720.0000000      105.9888151    -6645.7923980     3880.6765482        2.2122649        3.5384408        5.7118028
      1080.0000000     1489.9609958     6883.3690649     1288.9836882       -1.8270304        1.6789127       -7.1889009
      1440.0000000    -2412.4189938    -4550.2477242    -5617.3373135        0.2662198       -5.7477293        4.2337912


 1 34545U 93036UF  22068.39378412 +.00007551  00000 0  14486-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34545  74.0362  56.3228 0010568 112.5254 247.7033 14.6384882769013                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     3917.3009874     5878.8095585       -0.0040180       -1.7265468        1.1324576        7.2226438
       360.0000000     -788.3003797    -4078.4745555    -5720.8021080        4.4643235        4.5939784       -3.8999692
       720.0000000    -3092.5215443    -1363.5085127     6181.4905344       -3.2103305       -6.1320593       -2.9520881
      1080.0000000     4267.1413822     5526.3355213    -1088.1269107       -0.9828194        2.1454148        7.1292761
      1440.0000000    -1568.3970634    -4679.6419980    -5060.4350353        4.3688829        3.7354359       -4.8211006


 1 34546U 93036UG  22069.02028708 +.00000734  00000 0  25473-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34546  74.0328 270.3327 0024582 320.8666 192.7630 14.3719774968287                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000      837.1928488     6423.6534614     3054.2957125       -1.8550344        3.3001106       -6.4203555
       360.0000000      212.6733484    -7087.5985942      797.6949223        2.0452543        0.8604810        7.1460295
       720.0000000    -1192.3650466     5544.4496839    -4362.6847422       -1.6312039       -4.7257349       -5.5351467
      1080.0000000     1839.2230576    -2155.3289023     6560.2841965        0.7425306        7.1109012        2.1421374
      1440.0000000    -1944.1638974    -1829.8707332    -6620.0377302        0.3543331       -7.2167561        1.9021855


 1 34547U 93036UH  22068.73290787 +.00002426  00000 0  67758-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34547  74.0407 186.0122 0036889 249.0105 110.7105 14.4636143468524                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -7088.4823277     -746.5548439       -0.0025366        0.1970261       -2.0446782        7.1873314
       360.0000000     5223.8204410     1789.7083813    -4460.7142859       -5.0386561        1.0764255       -5.4527179
       720.0000000     -817.0086170    -2023.8435548     6791.7849906        7.3843973        0.3722612        0.9902070
      1080.0000000    -4051.2637499     1307.2139194    -5691.8642347       -6.1391634       -1.6329623        3.9677195
      1440.0000000     6864.7895778      -36.3517771     1861.3222308        1.8723223        2.1228163       -6.9347113


 1 34550U 93036UL  22068.48847750 +.00005213  00000 0  16483-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34550  73.9920 106.8624 0177754 242.5189 115.7794 14.3175270967438                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -2095.4422770     6913.2309044       -0.0110208       -1.9836406       -0.4861191        7.1129410
       360.0000000     2603.1722284    -5895.4238220    -2898.6473716        0.8691317        3.6909601       -6.5438452
       720.0000000    -2735.0255280     3606.2312669     5708.7670503        0.5138838       -6.0639519        4.0879770
      1080.0000000     2083.1158712     -279.7546017    -6732.3133673       -1.7576019        7.3082157       -0.9004358
      1440.0000000    -1210.3489496    -2773.3502821     6577.5899063        2.4650594       -6.5232602       -2.4016464


 1 34552U 93036UN  22069.14806022 +.00004264  00000 0  10926-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34552  74.0336 195.5332 0031631 222.1475 202.7766 14.5034212868439                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -2451.8058465    -2520.2484976     6187.9933983        6.7503579        0.9702670        3.0529161
       360.0000000    -2822.0145913     1142.5722758    -6415.4339138       -6.6205330       -2.5050359        2.4424259
       720.0000000     6395.8770846      788.7830289     2962.7297395        2.6754970        2.6179481       -6.5029283
      1080.0000000    -6362.4560750    -2268.8012343     2261.9484876        2.8666922       -1.2777616        6.7804405
      1440.0000000     2578.1750289     2434.8086443    -6138.1973912       -6.7732985       -0.7206671       -3.1430950


 1 34553U 93036UP  22069.20770096 +.00001545  00000 0  48719-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34553  74.0298 242.2505 0026659 293.3152 241.7380 14.4095706568278                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     3468.1199460     6216.3231838      610.7343394       -1.5052973        1.5314985       -7.1513109
       360.0000000    -1965.7259450    -5841.2445326     3593.8713171        3.3913186        2.6151560        6.1264927
       720.0000000     -285.9217919     3256.6074671    -6332.7423297       -4.0044578       -5.6935033       -2.7321612
      1080.0000000     2515.8039507      595.3669250     6661.1799069        3.0858398        6.5609330       -1.7455457
      1440.0000000    -3798.7757526    -4159.8758473    -4358.9724347       -0.9770850       -4.9314301        5.5464876


 1 34554U 93036UQ  22068.68224066 +.00000410  00000 0  15516-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34554  74.0390 300.0901 0024846 340.2587  75.5590 14.3416370568035                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     3412.7352734    -2636.9134673     5698.1809333       -2.1086656        5.9319285        4.0218355
       360.0000000    -1967.5501522     -513.6098325    -6863.0747194        3.5502561       -6.5361822       -0.5101946
       720.0000000      -19.1473686     3651.9042404     6162.0325354       -4.0490754        5.3858710       -3.1832236
      1080.0000000     1902.0268807    -5738.7738523    -3806.5954117        3.4886440       -2.8058525        5.9872533
      1440.0000000    -3293.5492160     6362.5179357      389.9562235       -1.9998729       -0.5949449       -7.1484084


 1 34555U 93036UR  22069.20221691 +.00004198  00000 0  97755-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34555  74.0268 137.2536 0021907 183.9846 241.7047 14.5501068368625                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -3369.7934555      697.4711487     6197.2531772        4.4324176       -5.2571674        2.9790622
       360.0000000     -959.0112743     3335.6385044    -6193.7502416       -5.5417314        4.0261779        3.0098218
       720.0000000     4529.5024545    -5084.9542682     1912.2685579        2.9012540       -0.0284049       -6.9345440
      1080.0000000    -5029.1558275     3424.7292901     3656.7982534        1.6790177       -4.0512245        6.0694855
      1440.0000000     2072.2362701      670.4191383    -6748.9242850       -5.0657089        5.4182524       -1.0344161


 1 34556U 93036US  22068.81781138 +.00002697  00000 0  77989-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34556  74.0310 214.3189 0022536 252.7941 224.2692 14.4488774168324                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     3651.6652722      375.1973987     6107.7170737        4.9745105        4.5242024       -3.2589439
       360.0000000    -5874.7114542    -3021.1155785    -2666.9990015       -1.3736224       -3.2008931        6.6196111
       720.0000000     5380.5004648     4220.7647291    -1967.1521431       -2.8897861        0.4567130       -6.8943398
      1080.0000000    -2395.4030389    -3505.3004119     5728.9994480        5.8528025        2.4469367        3.9409943
      1440.0000000    -1739.7121855     1198.4529526    -6792.0806043       -6.1462097       -4.1894868        0.8275098


 1 34560U 93036UW  22068.57206785 +.00002190  00000 0  64784-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34560  74.0278 222.2738 0029016 276.5018 155.9951 14.4379739368276                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     -317.7667203    -2822.7967181     6546.4348246        5.6816733        4.3341021        2.1498967
       360.0000000    -3151.6833158     -373.6654904    -6363.4928416       -4.6968413       -5.2086623        2.6264213
       720.0000000     5290.2630674     3350.7801951     3418.3087497        1.6123076        3.7925278       -6.2307321
      1080.0000000    -5106.7581282    -4844.0750014     1118.5051073        2.2419636       -0.7442573        7.0971210
      1440.0000000     2688.7836210     4199.3563099    -5077.0856586       -5.1300356       -2.5594684       -4.8146546


 1 34561U 93036UX  22068.81866329 +.00002883  00000 0  63805-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34561  74.0322  89.5651 0038908 166.5515 193.6696 14.5721106568734                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000       53.9657224     7109.5407563       -0.0005776       -2.0552285        0.0011865        7.1872294
       360.0000000     1457.6455845    -4412.9859307    -5334.9277129        1.3793470        5.8483191       -4.4983947
       720.0000000    -1937.0748456    -1391.5195200     6645.8724052        0.2493746       -7.3641022       -1.4940922
      1080.0000000     1092.1020551     6308.9972043    -3097.2276713       -1.7200508        3.4428817        6.4027515
      1440.0000000      541.8126429    -6458.8686912    -2809.4869869        2.0097025        3.0110068       -6.5934747


 1 34563U 93036UZ  22068.91980734 +.00000721  00000 0  23814-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34563  74.0392 232.3082 0015247 303.3138 183.2627 14.3977949368386                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     3853.9351573     2408.1840309     5511.2367821        2.6976240        5.4907041       -4.2769454
       360.0000000    -4668.3601756    -4952.8031868    -2136.1080746        0.0955972       -3.0398204        6.8332504
       720.0000000     3811.1205132     5696.7001558    -2017.3569485       -2.9045423       -0.4937826       -6.8621330
      1080.0000000    -1528.2067221    -4353.7069243     5443.9939628        4.7066182        3.8128281        4.3746097
      1440.0000000    -1322.6553498     1478.9313846    -6853.8452190       -4.8159565       -5.7021445       -0.2945313


 1 34564U 93036VA  22069.19372862 +.00000400  00000 0  14311-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34564  74.0322 250.9380 0020030 331.0174 151.3479 14.3720519868315                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     2825.2567142     3093.5823394     5797.6561457        1.0168246        6.3091889       -3.8449155
       360.0000000    -2941.0759409    -5833.1634089    -2873.5716546        0.7145677       -3.5739552        6.5317965
       720.0000000     2155.1428701     6756.5737361     -989.8946159       -2.2737053       -0.3169221       -7.0939754
      1080.0000000     -652.6674370    -5471.2631767     4537.7391858        3.1708512        4.0895208        5.3957549
      1440.0000000    -1066.0246689     2511.7749822    -6607.8199279       -3.0837660       -6.5042770       -1.9598227


 1 34567U 93036VD  22068.53031995 +.00000275  00000 0  13201-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34567  74.0407 105.3088 0075535 150.0956 329.2084 14.2195962067579                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     -747.7476356    -3785.3715232     6012.8520284        2.6942312       -6.0515938       -3.5020838
       360.0000000     -121.4242106     5602.0088206    -4610.5981918       -2.7221306        4.3301825        5.3203948
       720.0000000      937.5118019    -6599.2404342     2566.8179187        2.5515989       -2.2558760       -6.6825110
      1080.0000000    -1629.2303324     7049.7859083     -419.7212664       -2.0677873       -0.0897192        7.0971657
      1440.0000000     2163.9460248    -6523.9197219    -2024.4274272        1.4104655        2.5517825       -6.8852960


 1 34569U 93036VF  22068.95182238 +.00000620  00000 0  27037-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34569  73.9945 147.7061 0171535 287.6536 248.8497 14.1853831367451                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000     6012.0332640    -4014.5349959      632.5380343        1.5588663        1.4167318       -7.0828939
       360.0000000    -6133.1518239     3350.6624250     1737.6614619        0.3967162       -2.6258698        6.9561828
       720.0000000     5852.6154992    -2750.5538701    -3141.8658777       -1.9338720        3.4334780       -6.3307741
      1080.0000000    -4958.2132265     1547.1320959     5080.6088194        3.6251634       -4.1111869        4.9354854
      1440.0000000     3914.5470647     -594.5813596    -5926.9794407       -4.8524669        4.5116012       -3.5489888


 1 34570U 93036VG  22068.47157709 +.00000492  00000 0  32097-3 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34570  73.9598  78.8025 0201588 249.6357 264.3030 13.9562621166403                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -2186.9435677    -6124.6596991     3322.9351673        1.1239429       -3.6593768       -6.3104142
       360.0000000     2223.8518556     6078.2678509    -3312.5739335       -1.0129599        3.9254414        6.2071093
       720.0000000    -2413.6393734    -5459.7254403     4239.2570159        0.7043060       -4.5745362       -5.7237133
      1080.0000000     2432.8079096     5377.3010289    -4213.1581343       -0.5769029        4.8439817        5.6071082
      1440.0000000    -2562.2202628    -4661.9693486     5056.3671253        0.2455217       -5.3633010       -5.0075791


 1 34571U 93036VH  22067.56494129 +.00002866  00000 0  12100-2 0 0999                                                                                                                                                                                                                                                                                                                                                                                                                                                           
 2 34571  74.0520 174.2016 0101176 188.3648 171.5817 14.2332293367699                                                                                                                                                                                                                                                                                                                                                                                                                                                           

         0.0000000    -7229.7017024      734.1633599       -0.0035311       -0.2079841       -2.0133352        7.0874727
       360.0000000     6697.9834448      -33.5464928    -2432.4243764       -2.4217747        2.2081029       -6.7632129
       720.0000000    -5650.1010709     -631.8419903     4498.7000233        4.5912232       -2.0897964        5.3931204
      1080.0000000     3506.6638702     1317.9263360    -6098.6379352       -6.4254660        1.7795507       -3.3841634
      1440.0000000    -1406.8304520    -1784.4717301     6836.0698389        7.2370171       -1.2753884        1.0777075
Total run time =       0.03 seconds.
"""